# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""The tests module contains the tests of the packages/contracts/erc1155 dir."""

# type: ignore # noqa: E800
# pylint: skip-file

import re
import time
from pathlib import Path
from typing import cast
from unittest import mock

import pytest
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_ethereum.test_tools.constants import (
    ETHEREUM_ADDRESS_ONE,
    ETHEREUM_ADDRESS_TWO,
    ETHEREUM_PRIVATE_KEY_PATH,
    ETHEREUM_PRIVATE_KEY_TWO_PATH,
    ETHEREUM_TESTNET_CONFIG,
)
from aea_ledger_ethereum.test_tools.fixture_helpers import (  # noqa  # pylint: disable=unsed-import
    ganache,
)
from aea_ledger_fetchai import FetchAIApi, FetchAICrypto
from aea_ledger_fetchai.test_tools.constants import FETCHAI_TESTNET_CONFIG

from aea.configurations.loader import (
    ComponentType,
    ContractConfig,
    load_component_configuration,
)
from aea.contracts.base import Contract, contract_registry
from aea.test_tools.test_contract import BaseContractTestCase

PACKAGE_DIR = Path(__file__).parent.parent
MAX_FLAKY_RERUNS = 3


@pytest.mark.ledger
@pytest.mark.integration
@pytest.mark.usefixtures("ganache")
class TestERC1155ContractEthereum(BaseContractTestCase):
    """Test the ERC1155 contract on Ethereum."""

    ledger_identifier = EthereumCrypto.identifier
    path_to_contract = PACKAGE_DIR

    @classmethod
    def setup_method(cls) -> None:
        """Setup."""
        super().setup_method(
            ledger_config=ETHEREUM_TESTNET_CONFIG,
            deployer_private_key_path=ETHEREUM_PRIVATE_KEY_PATH,
            item_owner_private_key_path=ETHEREUM_PRIVATE_KEY_TWO_PATH,
        )

        cls.token_ids_a = [
            340282366920938463463374607431768211456,
            340282366920938463463374607431768211457,
            340282366920938463463374607431768211458,
            340282366920938463463374607431768211459,
            340282366920938463463374607431768211460,
            340282366920938463463374607431768211461,
            340282366920938463463374607431768211462,
            340282366920938463463374607431768211463,
            340282366920938463463374607431768211464,
            340282366920938463463374607431768211465,
        ]

        cls.token_id_b = 680564733841876926926749214863536422912

    @classmethod
    def finish_contract_deployment(cls) -> str:
        """
        Finish deploying contract.

        :return: contract address
        """
        contract_address = cls.ledger_api.get_contract_address(
            cls.deployment_tx_receipt
        )

        if contract_address is None:
            raise ValueError("Contract address not found!")  # pragma: nocover

        return contract_address

    def test_generate_token_ids(self) -> None:
        """Test the generate_token_ids method of the ERC1155 contract."""
        # setup
        nft_token_type = 1
        nb_tokens = 2
        expected_toke_ids = [
            340282366920938463463374607431768211456,
            340282366920938463463374607431768211457,
        ]

        # operation
        actual_toke_ids = self.contract.generate_token_ids(nft_token_type, nb_tokens)

        # after
        assert actual_toke_ids == expected_toke_ids

    def test_generate_id(self) -> None:
        """Test the _generate_id method of the ERC1155 contract."""
        # setup
        ft_token_type = 2
        index = 0
        expected_toke_id = 680564733841876926926749214863536422912

        # operation
        actual_toke_id = self.contract._generate_id(index, ft_token_type)

        # after
        assert actual_toke_id == expected_toke_id

    def test_get_create_batch_transaction(self) -> None:
        """Test the get_create_batch_transaction method of the ERC1155 contract."""
        # operation
        tx = self.contract.get_create_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            token_ids=self.token_ids_a,
        )

        # after
        assert len(tx) == 7
        assert all(
            key in tx
            for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to", "data"]
        )
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

    def test_get_create_single_transaction(self) -> None:
        """Test the get_create_single_transaction method of the ERC1155 contract."""
        # operation
        tx = self.contract.get_create_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            token_id=self.token_id_b,
        )

        # after
        assert len(tx) == 7
        assert all(
            key in tx
            for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to", "data"]
        )
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

    def test_get_mint_batch_transaction(self) -> None:
        """Test the get_mint_batch_transaction method of the ERC1155 contract."""
        # operation
        tx = self.contract.get_mint_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.item_owner_crypto.address,
            token_ids=self.token_ids_a,
            mint_quantities=[1] * len(self.token_ids_a),
        )

        # after
        assert len(tx) == 7
        assert all(
            key in tx
            for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to", "data"]
        )
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

    def test_validate_mint_quantities(self) -> None:
        """Test the validate_mint_quantities method of the ERC1155 contract."""
        # Valid NFTs
        self.contract.validate_mint_quantities(
            token_ids=self.token_ids_a,
            mint_quantities=[1] * len(self.token_ids_a),
        )

        # Valid FTs
        token_id = 680564733841876926926749214863536422912
        mint_quantity = 1
        self.contract.validate_mint_quantities(
            token_ids=[token_id],
            mint_quantities=[mint_quantity],
        )

        # Invalid NFTs
        token_id = self.token_ids_a[0]
        mint_quantity = 2
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Cannot mint NFT (token_id={token_id}) with mint_quantity more than 1 (found={mint_quantity})"
            ),
        ):
            self.contract.validate_mint_quantities(
                token_ids=[token_id],
                mint_quantities=[mint_quantity],
            )

        # Invalid: neither NFT nor FT
        token_id = 1020847100762815390390123822295304634368
        mint_quantity = 1
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"The token type must be 1 or 2. Found type=3 for token_id={token_id}"
            ),
        ):
            self.contract.validate_mint_quantities(
                token_ids=[token_id],
                mint_quantities=[mint_quantity],
            )

    def test_decode_id(self) -> None:
        """Test the decode_id method of the ERC1155 contract."""
        # FT
        expected_token_type = 2
        token_id = 680564733841876926926749214863536422912
        actual_token_type = self.contract.decode_id(token_id)
        assert actual_token_type == expected_token_type

        # NFT
        expected_token_type = 1
        token_id = 340282366920938463463374607431768211456
        actual_token_type = self.contract.decode_id(token_id)
        assert actual_token_type == expected_token_type

    def test_get_mint_single_transaction(self) -> None:
        """Test the get_mint_single_transaction method of the ERC1155 contract."""
        # operation
        tx = self.contract.get_mint_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.item_owner_crypto.address,
            token_id=self.token_id_b,
            mint_quantity=1,
        )

        # after
        assert len(tx) == 7
        assert all(
            key in tx
            for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to", "data"]
        )
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

    def test_get_balance(self) -> None:
        """Test the get_balance method of the ERC1155 contract."""
        # operation
        result = self.contract.get_balance(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.item_owner_crypto.address,
            token_id=self.token_id_b,
        )

        # after
        assert "balance" in result
        assert result["balance"][self.token_id_b] == 0

    def test_get_balances(self) -> None:
        """Test the get_balances method of the ERC1155 contract."""
        # operation
        result = self.contract.get_balances(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.item_owner_crypto.address,
            token_ids=self.token_ids_a,
        )

        # after
        assert "balances" in result
        assert all(result["balances"][token_id] == 0 for token_id in self.token_ids_a)

    def test_get_hash_single(self) -> None:
        """Test the get_hash_single method of the ERC1155 contract."""
        # operation
        result = self.contract.get_hash_single(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            from_address=self.deployer_crypto.address,
            to_address=self.item_owner_crypto.address,
            token_id=self.token_id_b,
            from_supply=0,
            to_supply=10,
            value=1,
            trade_nonce=1,
        )

        # after
        assert isinstance(result, bytes)

    def test_get_hash_batch(self) -> None:
        """Test the get_hash_batch method of the ERC1155 contract."""
        # operation
        result = self.contract.get_hash_batch(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            from_address=self.deployer_crypto.address,
            to_address=self.item_owner_crypto.address,
            token_ids=self.token_ids_a,
            from_supplies=[0, 1, 0, 0, 1, 0, 0, 0, 0, 1],
            to_supplies=[0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            value=1,
            trade_nonce=1,
        )

        # after
        assert isinstance(result, bytes)

    def test_generate_trade_nonce(self) -> None:
        """Test the generate_trade_nonce method of the ERC1155 contract."""
        # operation
        result = self.contract.generate_trade_nonce(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.item_owner_crypto.address,
        )

        # after
        assert "trade_nonce" in result
        assert isinstance(result["trade_nonce"], int)

    @pytest.mark.integration
    def test_helper_methods_and_get_transactions(self) -> None:
        """Test helper methods and get transactions."""
        expected_a = [
            340282366920938463463374607431768211456,
            340282366920938463463374607431768211457,
            340282366920938463463374607431768211458,
            340282366920938463463374607431768211459,
            340282366920938463463374607431768211460,
            340282366920938463463374607431768211461,
            340282366920938463463374607431768211462,
            340282366920938463463374607431768211463,
            340282366920938463463374607431768211464,
            340282366920938463463374607431768211465,
        ]
        actual = self.contract.generate_token_ids(token_type=1, nb_tokens=10)
        assert expected_a == actual
        expected_b = [
            680564733841876926926749214863536422912,
            680564733841876926926749214863536422913,
        ]
        actual = self.contract.generate_token_ids(token_type=2, nb_tokens=2)
        assert expected_b == actual
        tx = self.contract.get_deploy_transaction(
            ledger_api=self.ledger_api, deployer_address=ETHEREUM_ADDRESS_ONE
        )
        assert len(tx) == 7
        data = tx.pop("data")
        assert len(data) > 0 and data.startswith("0x")
        assert all(
            [key in tx for key in ["value", "from", "gas", "gasPrice", "nonce"]]
        ), "Error, found: {}".format(tx)
        tx = self.contract.get_create_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=ETHEREUM_ADDRESS_ONE,
            deployer_address=ETHEREUM_ADDRESS_ONE,
            token_ids=expected_a,
        )
        assert len(tx) == 7
        data = tx.pop("data")
        assert len(data) > 0 and data.startswith("0x")
        assert all(
            [
                key in tx
                for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to"]
            ]
        ), "Error, found: {}".format(tx)
        tx = self.contract.get_create_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=ETHEREUM_ADDRESS_ONE,
            deployer_address=ETHEREUM_ADDRESS_ONE,
            token_id=expected_b[0],
        )
        assert len(tx) == 7
        data = tx.pop("data")
        assert len(data) > 0 and data.startswith("0x")
        assert all(
            [
                key in tx
                for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to"]
            ]
        ), "Error, found: {}".format(tx)
        mint_quantities = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        tx = self.contract.get_mint_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=ETHEREUM_ADDRESS_ONE,
            deployer_address=ETHEREUM_ADDRESS_ONE,
            recipient_address=ETHEREUM_ADDRESS_ONE,
            token_ids=expected_a,
            mint_quantities=mint_quantities,
        )
        assert len(tx) == 7
        data = tx.pop("data")
        assert len(data) > 0 and data.startswith("0x")
        assert all(
            [
                key in tx
                for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to"]
            ]
        ), "Error, found: {}".format(tx)
        mint_quantity = 1
        tx = self.contract.get_mint_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=ETHEREUM_ADDRESS_ONE,
            deployer_address=ETHEREUM_ADDRESS_ONE,
            recipient_address=ETHEREUM_ADDRESS_ONE,
            token_id=expected_b[1],
            mint_quantity=mint_quantity,
        )
        assert len(tx) == 7
        data = tx.pop("data")
        assert len(data) > 0 and data.startswith("0x")
        assert all(
            [
                key in tx
                for key in ["value", "chainId", "gas", "gasPrice", "nonce", "to"]
            ]
        ), "Error, found: {}".format(tx)

    @pytest.mark.integration
    def test_get_single_atomic_swap(self) -> None:
        """Test get single atomic swap."""
        from_address = ETHEREUM_ADDRESS_ONE
        to_address = ETHEREUM_ADDRESS_TWO
        token_id = self.contract.generate_token_ids(token_type=2, nb_tokens=1)[0]
        from_supply = 0
        to_supply = 10
        value = 1
        trade_nonce = 1
        tx_hash = self.contract.get_hash_single(
            self.ledger_api,
            self.contract_address,
            from_address,
            to_address,
            token_id,
            from_supply,
            to_supply,
            value,
            trade_nonce,
        )
        assert isinstance(tx_hash, bytes)
        signature = self.deployer_crypto.sign_message(tx_hash)
        tx = self.contract.get_atomic_swap_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            from_address=from_address,
            to_address=to_address,
            token_id=token_id,
            from_supply=from_supply,
            to_supply=to_supply,
            value=value,
            trade_nonce=trade_nonce,
            signature=signature,
        )
        assert len(tx) == 8
        data = tx.pop("data")
        assert len(data) > 0 and data.startswith("0x")
        assert all(
            [
                key in tx
                for key in [
                    "value",
                    "chainId",
                    "gas",
                    "gasPrice",
                    "nonce",
                    "to",
                    "from",
                ]
            ]
        ), "Error, found: {}".format(tx)

        with pytest.raises(
            RuntimeError,
            match="Can't determine direction of swap because from_supply and to_supply are both non-zero",
        ):
            self.contract.get_atomic_swap_single_transaction(
                ledger_api=self.ledger_api,
                contract_address=self.contract_address,
                from_address=from_address,
                to_address=to_address,
                token_id=token_id,
                from_supply=10,
                to_supply=10,
                value=value,
                trade_nonce=trade_nonce,
                signature=signature,
            )

    @pytest.mark.integration
    def test_get_batch_atomic_swap(self) -> None:
        """Test get batch atomic swap."""
        from_address = ETHEREUM_ADDRESS_ONE
        to_address = ETHEREUM_ADDRESS_TWO
        token_ids = self.contract.generate_token_ids(token_type=2, nb_tokens=10)
        from_supplies = [0, 1, 0, 0, 1, 0, 0, 0, 0, 1]
        to_supplies = [0, 0, 0, 0, 0, 1, 0, 0, 0, 0]
        value = 1
        trade_nonce = 1
        tx_hash = self.contract.get_hash_batch(
            self.ledger_api,
            self.contract_address,
            from_address,
            to_address,
            token_ids,
            from_supplies,
            to_supplies,
            value,
            trade_nonce,
        )
        assert isinstance(tx_hash, bytes)
        signature = self.deployer_crypto.sign_message(tx_hash)
        tx = self.contract.get_atomic_swap_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            from_address=from_address,
            to_address=to_address,
            token_ids=token_ids,
            from_supplies=from_supplies,
            to_supplies=to_supplies,
            value=value,
            trade_nonce=trade_nonce,
            signature=signature,
        )
        assert len(tx) == 8
        data = tx.pop("data")
        assert len(data) > 0 and data.startswith("0x")
        assert all(
            [
                key in tx
                for key in [
                    "value",
                    "chainId",
                    "gas",
                    "gasPrice",
                    "nonce",
                    "to",
                    "from",
                ]
            ]
        ), "Error, found: {}".format(tx)

    @pytest.mark.integration
    def test_full(self) -> None:
        """Setup."""
        # Test tokens IDs
        token_ids = self.contract.generate_token_ids(token_type=2, nb_tokens=10)

        # create
        tx = self.contract.get_create_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            token_ids=token_ids,
        )
        tx_signed = self.deployer_crypto.sign_transaction(tx)
        tx_receipt = self.ledger_api.send_signed_transaction(tx_signed)
        time.sleep(1)
        receipt = self.ledger_api.get_transaction_receipt(tx_receipt)
        assert self.ledger_api.is_transaction_settled(receipt)

        mint_quantities = [10] * len(token_ids)
        # mint
        tx = self.contract.get_mint_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.deployer_crypto.address,
            token_ids=token_ids,
            mint_quantities=mint_quantities,
        )
        tx_signed = self.deployer_crypto.sign_transaction(tx)
        tx_receipt = self.ledger_api.send_signed_transaction(tx_signed)
        time.sleep(1)
        receipt = self.ledger_api.get_transaction_receipt(tx_receipt)
        assert self.ledger_api.is_transaction_settled(receipt)

        tx = self.contract.get_mint_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.item_owner_crypto.address,
            token_ids=token_ids,
            mint_quantities=mint_quantities,
        )
        tx_signed = self.deployer_crypto.sign_transaction(tx)
        tx_receipt = self.ledger_api.send_signed_transaction(tx_signed)
        time.sleep(1)
        receipt = self.ledger_api.get_transaction_receipt(tx_receipt)
        assert self.ledger_api.is_transaction_settled(receipt)

        #  batch trade
        from_address = self.deployer_crypto.address
        to_address = self.item_owner_crypto.address
        from_supplies = [0, 1, 0, 0, 1, 0, 0, 0, 0, 1]
        to_supplies = [0, 0, 0, 0, 0, 1, 0, 0, 0, 0]
        value = 0
        trade_nonce = 1
        tx_hash = self.contract.get_hash_batch(
            self.ledger_api,
            self.contract_address,
            from_address,
            to_address,
            token_ids,
            from_supplies,
            to_supplies,
            value,
            trade_nonce,
        )
        signature = self.item_owner_crypto.sign_message(
            tx_hash, is_deprecated_mode=True
        )
        tx = self.contract.get_atomic_swap_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            from_address=from_address,
            to_address=to_address,
            token_ids=token_ids,
            from_supplies=from_supplies,
            to_supplies=to_supplies,
            value=value,
            trade_nonce=trade_nonce,
            signature=signature,
        )
        tx_signed = self.deployer_crypto.sign_transaction(tx)
        tx_receipt = self.ledger_api.send_signed_transaction(tx_signed)
        time.sleep(1)
        receipt = self.ledger_api.get_transaction_receipt(tx_receipt)
        assert self.ledger_api.is_transaction_settled(receipt)


@pytest.mark.skip("Fetch.ai testnet not currently working.")
class TestCosmWasmContract(BaseContractTestCase):
    """Test the cosmwasm contract."""

    ledger_identifier = FetchAICrypto.identifier
    path_to_contract = PACKAGE_DIR
    fund_from_faucet = True

    @classmethod
    def setup_class(cls) -> None:
        """Setup."""
        # Test tokens IDs
        super().setup_class(ledger_config=FETCHAI_TESTNET_CONFIG)
        cls.token_ids_a = [
            340282366920938463463374607431768211456,
            340282366920938463463374607431768211457,
            340282366920938463463374607431768211458,
            340282366920938463463374607431768211459,
            340282366920938463463374607431768211460,
            340282366920938463463374607431768211461,
            340282366920938463463374607431768211462,
            340282366920938463463374607431768211463,
            340282366920938463463374607431768211464,
            340282366920938463463374607431768211465,
        ]

        cls.token_id_b = 680564733841876926926749214863536422912

    @classmethod
    def finish_contract_deployment(cls) -> str:
        """
        Finish deploying contract.

        :return: contract address
        """
        code_id = cast(FetchAIApi, cls.ledger_api).get_code_id(
            cls.deployment_tx_receipt
        )

        assert code_id is not None

        # Init contract
        tx = cls._contract.get_deploy_transaction(
            ledger_api=cls.ledger_api,
            deployer_address=cls.deployer_crypto.address,
            code_id=code_id,
            init_msg={},
            tx_fee=0,
            amount=0,
            label="ERC1155",
            gas=1000000,
        )

        if tx is None:
            raise ValueError("Deploy transaction not found!")  # pragma: nocover

        tx_receipt = cls.sign_send_confirm_receipt_multisig_transaction(
            tx, cls.ledger_api, [cls.deployer_crypto]
        )

        contract_address = cls.ledger_api.get_contract_address(tx_receipt)

        if contract_address is None:
            raise ValueError("Contract address not found!")  # pragma: nocover

        return contract_address

    @pytest.mark.integration
    @pytest.mark.ledger
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    def test_create_and_mint_and_balances(self) -> None:
        """Test cosmwasm contract create, mint and balances functionalities."""
        # Create single token
        tx = self.contract.get_create_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            token_id=self.token_id_b,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Create batch of tokens
        tx = self.contract.get_create_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            token_ids=self.token_ids_a,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Mint single token
        tx = self.contract.get_mint_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.item_owner_crypto.address,
            token_id=self.token_id_b,
            mint_quantity=1,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Get balance of single token
        res = self.contract.get_balance(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.item_owner_crypto.address,
            token_id=self.token_id_b,
        )
        assert "balance" in res
        assert res["balance"][self.token_id_b] == 1

        # Mint batch of tokens
        tx = self.contract.get_mint_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.item_owner_crypto.address,
            token_ids=self.token_ids_a,
            mint_quantities=[1] * len(self.token_ids_a),
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Get balances of multiple tokens
        res = self.contract.get_balances(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.item_owner_crypto.address,
            token_ids=self.token_ids_a,
        )
        assert "balances" in res
        assert res["balances"] == {token_id: 1 for token_id in self.token_ids_a}

    @pytest.mark.integration
    @pytest.mark.ledger
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    def test_cosmwasm_single_atomic_swap(self) -> None:
        """Test single atomic swap."""
        # Create batch of tokens
        tx = self.contract.get_create_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            token_ids=self.token_ids_a,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Mint single ERC1155 token a[0] to Deployer
        tx = self.contract.get_mint_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.deployer_crypto.address,
            token_id=self.token_ids_a[0],
            mint_quantity=1,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Store balance of Deployer's native tokens before atomic swap
        original_deployer_balance = self.ledger_api.get_balance(
            self.deployer_crypto.address
        )

        # Atomic swap
        # Send 1 ERC1155 token a[0] from Deployer to Item owner
        # Send 1 native token from Item owner to Deployer
        tx = self.contract.get_atomic_swap_single_transaction(
            self.ledger_api,
            contract_address=self.contract_address,
            from_address=self.deployer_crypto.address,
            to_address=self.item_owner_crypto.address,
            token_id=self.token_ids_a[0],
            from_supply=1,
            to_supply=0,
            value=1,
            trade_nonce=0,
            from_pubkey=self.deployer_crypto.public_key,
            to_pubkey=self.item_owner_crypto.public_key,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto, self.item_owner_crypto]
        )

        # Check Item owner's ERC1155 token balance
        result = self.contract.get_balance(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.item_owner_crypto.address,
            token_id=self.token_ids_a[0],
        )

        assert "balance" in result
        assert result["balance"][self.token_ids_a[0]] == 1

        # Check deployer's native token balance
        deployer_balance = self.ledger_api.get_balance(self.deployer_crypto.address)
        assert deployer_balance == original_deployer_balance + 1

        # Other direction of atomic swap
        # Send 1 ERC1155 token a[0] from Item owner to Deployer
        # Send 1 native token from Item owner to Deployer
        tx = self.contract.get_atomic_swap_single_transaction(
            self.ledger_api,
            contract_address=self.contract_address,
            from_address=self.deployer_crypto.address,
            to_address=self.item_owner_crypto.address,
            token_id=self.token_ids_a[0],
            from_supply=0,
            to_supply=1,
            value=1,
            trade_nonce=0,
            from_pubkey=self.deployer_crypto.public_key,
            to_pubkey=self.item_owner_crypto.public_key,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.item_owner_crypto]
        )

        # Check Item owner's ERC1155 token balance
        result = self.contract.get_balance(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.deployer_crypto.address,
            token_id=self.token_ids_a[0],
        )

        assert "balance" in result
        assert result["balance"][self.token_ids_a[0]] == 1

        # Check deployer's native token balance
        deployer_balance = self.ledger_api.get_balance(self.deployer_crypto.address)
        assert deployer_balance == original_deployer_balance + 2

        # Check invalid case with from_supply > 0 and to_supply > 0
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address=self.contract_address,
                from_address=self.deployer_crypto.address,
                to_address=self.item_owner_crypto.address,
                token_id=self.token_ids_a[0],
                from_supply=1,
                to_supply=1,
                value=1,
                trade_nonce=0,
                from_pubkey=self.deployer_crypto.public_key,
                to_pubkey=self.item_owner_crypto.public_key,
            )

    @pytest.mark.integration
    @pytest.mark.ledger
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    def test_cosmwasm_batch_atomic_swap(self) -> None:
        """Test batch atomic swap."""

        # Create batch of tokens
        tx = self.contract.get_create_batch_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            token_ids=self.token_ids_a,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Mint single token a[0] to Deployer
        tx = self.contract.get_mint_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.deployer_crypto.address,
            token_id=self.token_ids_a[0],
            mint_quantity=1,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Mint single token a[1] to Item owner
        tx = self.contract.get_mint_single_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            deployer_address=self.deployer_crypto.address,
            recipient_address=self.item_owner_crypto.address,
            token_id=self.token_ids_a[1],
            mint_quantity=1,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto]
        )

        # Store balance of Deployer's native tokens before atomic swap
        original_deployer_balance = self.ledger_api.get_balance(
            self.deployer_crypto.address
        )

        # Atomic swap
        # Send 1 ERC1155 token a[0] from Deployer to Item owner
        # Send 1 ERC1155 token a[1] from Item owner to Deployer
        # Send 1 native token from Item owner to Deployer

        tx = self.contract.get_atomic_swap_batch_transaction(
            self.ledger_api,
            contract_address=self.contract_address,
            from_address=self.deployer_crypto.address,
            to_address=self.item_owner_crypto.address,
            token_ids=[self.token_ids_a[0], self.token_ids_a[1]],
            from_supplies=[1, 0],
            to_supplies=[0, 1],
            value=1,
            trade_nonce=0,
            from_pubkey=self.deployer_crypto.public_key,
            to_pubkey=self.item_owner_crypto.public_key,
        )
        assert len(tx) == 2
        self.sign_send_confirm_receipt_multisig_transaction(
            tx, self.ledger_api, [self.deployer_crypto, self.item_owner_crypto]
        )

        # Check Item owner's ERC1155 token balance
        result = self.contract.get_balance(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.item_owner_crypto.address,
            token_id=self.token_ids_a[0],
        )

        assert "balance" in result
        assert result["balance"][self.token_ids_a[0]] == 1

        # Check Deployer's ERC1155 token balance
        result = self.contract.get_balance(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            agent_address=self.deployer_crypto.address,
            token_id=self.token_ids_a[1],
        )

        assert "balance" in result
        assert result["balance"][self.token_ids_a[1]] == 1

        # Check deployer's native token balance
        deployer_balance = self.ledger_api.get_balance(self.deployer_crypto.address)
        assert deployer_balance == original_deployer_balance + 1


class TestContractCommon:
    """Other tests for the contract."""

    @classmethod
    def setup_class(cls) -> None:
        """Setup class."""

        # Register smart contract used for testing
        cls.path_to_contract = PACKAGE_DIR

        # register contract
        configuration = cast(
            ContractConfig,
            load_component_configuration(ComponentType.CONTRACT, cls.path_to_contract),
        )
        configuration._directory = (  # pylint: disable=protected-access
            cls.path_to_contract
        )
        if str(configuration.public_id) not in contract_registry.specs:
            # load contract into sys modules
            Contract.from_config(configuration)
        cls.contract = contract_registry.make(str(configuration.public_id))

    def setup_method(self) -> None:
        """Setup test method."""
        self.token_ids_a = [
            340282366920938463463374607431768211456,
        ]

        # Create mock ledger with unknown identifier
        self.ledger_api = mock.Mock()
        attrs = {"identifier": "dummy"}
        self.ledger_api.configure_mock(**attrs)

    @pytest.mark.ledger
    def test_get_create_batch_transaction_wrong_identifier(self) -> None:
        """Test if get_create_batch_transaction with wrong api identifier fails."""

        # Test if function is not implemented for unknown ledger
        with pytest.raises(NotImplementedError):
            self.contract.get_create_batch_transaction(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                deployer_address="address",
                token_ids=self.token_ids_a,
            )

    @pytest.mark.ledger
    def test_get_create_single_transaction_wrong_identifier(self) -> None:
        """Test if get_create_single_transaction with wrong api identifier fails."""

        # Test if function is not implemented for unknown ledger
        with pytest.raises(NotImplementedError):
            self.contract.get_create_single_transaction(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                deployer_address="address",
                token_id=self.token_ids_a[0],
            )

    @pytest.mark.ledger
    def test_get_mint_batch_transaction_wrong_identifier(self) -> None:
        """Test if get_mint_batch_transaction with wrong api identifier fails."""

        # Test if function is not implemented for unknown ledger
        with pytest.raises(NotImplementedError):
            self.contract.get_mint_batch_transaction(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                deployer_address="address",
                recipient_address="address",
                token_ids=self.token_ids_a,
                mint_quantities=[1],
            )

    @pytest.mark.ledger
    def test_get_mint_single_transaction_wrong_identifier(self) -> None:
        """Test if get_mint_single_transaction with wrong api identifier fails."""

        # Test if function is not implemented for unknown ledger
        with pytest.raises(NotImplementedError):
            self.contract.get_mint_single_transaction(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                deployer_address="address",
                recipient_address="address",
                token_id=self.token_ids_a[0],
                mint_quantity=1,
            )

    @pytest.mark.ledger
    def test_get_balance_wrong_identifier(self) -> None:
        """Test if get_balance with wrong api identifier fails."""

        # Test if function is not implemented for unknown ledger
        with pytest.raises(NotImplementedError):
            self.contract.get_balance(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                agent_address="address",
                token_id=self.token_ids_a[0],
            )

    @pytest.mark.ledger
    def test_get_balance_wrong_query_res(self) -> None:
        """Test if get_balance with wrong api identifier fails."""

        # Create mock fetchai ledger that returns None on execute_contract_query
        attrs = {"identifier": "fetchai", "execute_contract_query.return_value": None}
        self.ledger_api.configure_mock(**attrs)

        # Test if get balance returns ValueError when querying contract returns None
        with pytest.raises(ValueError):
            self.contract.get_balance(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                agent_address="address",
                token_id=self.token_ids_a[0],
            )

    @pytest.mark.ledger
    def test_get_balances_wrong_query_res(self) -> None:
        """Test if get_balances with wrong api identifier fails."""

        # Create mock fetchai ledger that returns None on execute_contract_query
        attrs = {"identifier": "fetchai", "execute_contract_query.return_value": None}
        self.ledger_api.configure_mock(**attrs)

        # Test if get balance returns ValueError when querying contract returns None
        with pytest.raises(ValueError):
            self.contract.get_balances(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                agent_address="address",
                token_ids=self.token_ids_a,
            )

    @pytest.mark.ledger
    def test_get_hash_batch_not_same(self) -> None:
        """Test if get_hash_batch returns ValueError when on-chain hash is not same as computed hash."""

        self.ledger_api.identifier = "ethereum"

        # Test if get hash returns ValueError when on chain hash is not same as computed hash
        with mock.patch.object(type(self.contract), "_get_hash_batch", new=mock.Mock()):
            with pytest.raises(ValueError):
                self.contract.get_hash_batch(
                    ledger_api=self.ledger_api,
                    contract_address="contract_address",
                    from_address="address",
                    to_address="address",
                    token_ids=self.token_ids_a,
                    from_supplies=[1],
                    to_supplies=[0],
                    value=123,
                    trade_nonce=123,
                )

    @pytest.mark.ledger
    def test_generate_trade_nonce_if_exist(self) -> None:
        """Test if generate_trade_nonce retries when nonce already exist."""

        # Etherem ledger api mock
        self.ledger_api.identifier = "ethereum"

        # instance.functions.is_nonce_used(agent_address, trade_nonce).call() -> True, False
        is_nonce_used_mock = mock.Mock()
        is_nonce_used_mock.configure_mock(**{"call.side_effect": [True, False]})

        # instance.functions.is_nonce_used(agent_address, trade_nonce) -> is_nonce_used_mock with call method
        instance_mock = mock.Mock()
        instance_mock.configure_mock(
            **{"functions.is_nonce_used.return_value": is_nonce_used_mock}
        )

        # cls.get_instance(ledger_api, contract_address) -> instance_mock
        get_instance_mock = mock.Mock()
        get_instance_mock.configure_mock(**{"return_value": instance_mock})

        # Patch get_instance method to return get_instance_mock which returns instance of instance_mock when called
        with mock.patch.object(
            type(self.contract), "get_instance", new=get_instance_mock
        ):
            self.contract.generate_trade_nonce(
                ledger_api=self.ledger_api,
                contract_address="contract_address",
                agent_address="address",
            )

        # Check if is_nonce_used was called twice
        assert is_nonce_used_mock.call.call_count == 2

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_eth_no_signature(self) -> None:
        """Test if get_atomic_swap_single_transaction returns RuntimeError if signature not present on Ethereum case."""

        self.ledger_api.identifier = "ethereum"

        # Test if get_atomic_swap_single_transaction returns RuntimeError when signature is missing
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=1,
                to_supply=0,
                value=1,
                trade_nonce=0,
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_eth_pubkeys(self) -> None:
        """Test if get_atomic_swap_single_transaction returns RuntimeError if pubkeys are present on Ethereum case."""

        self.ledger_api.identifier = "ethereum"

        # Test if get_atomic_swap_single_transaction returns RuntimeError when pubkey is present
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=1,
                to_supply=0,
                value=1,
                trade_nonce=0,
                signature="signature",
                from_pubkey="deadbeef",
                to_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_cosmos_signature(self) -> None:
        """Test if get_atomic_swap_single_transaction returns RuntimeError if signature is present on Cosmos/Fetch case."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction returns RuntimeError when signature is present
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=1,
                to_supply=0,
                value=1,
                trade_nonce=0,
                signature="signature",
                from_pubkey="deadbeef",
                to_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_cosmos_one_pubkey_valid(self) -> None:
        """Test if get_atomic_swap_single_transaction allows one pubkey in case of only one direction of transfers."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction works with only to_pubkey
        tx = self.contract.get_atomic_swap_single_transaction(
            self.ledger_api,
            contract_address="address",
            from_address="address",
            to_address="address",
            token_id=self.token_ids_a[0],
            from_supply=0,
            to_supply=1,
            value=1,
            trade_nonce=0,
            to_pubkey="deadbeef",
        )
        assert tx is not None

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_cosmos_one_pubkey_invalid(self) -> None:
        """Test if get_atomic_swap_single_transaction returns RuntimeError with missing from_pubkey."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with missing from_key
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=1,
                to_supply=0,
                value=1,
                trade_nonce=0,
                to_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_cosmos_to_pubkey_missing(self) -> None:
        """Test if get_atomic_swap_single_transaction returns RuntimeError with missing to_pubkey."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with missing from_key
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=1,
                to_supply=0,
                value=1,
                trade_nonce=0,
                from_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_eth_pubkeys(self) -> None:
        """Test if get_atomic_swap_batch_transaction returns RuntimeError if pubkeys are present on Ethereum case."""

        self.ledger_api.identifier = "ethereum"

        # Test if get_atomic_swap_batch_transaction returns RuntimeError when pubkey is present
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[1],
                to_supplies=[0],
                value=1,
                trade_nonce=0,
                signature="signature",
                from_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_cosmos_signature(self) -> None:
        """Test if get_atomic_swap_batch_transaction returns RuntimeError if signature is present on Cosmos/Fetch case."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_batch_transaction returns RuntimeError when signature is present
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[1],
                to_supplies=[0],
                value=1,
                trade_nonce=0,
                signature="signature",
                from_pubkey="deadbeef",
                to_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_cosmos_one_pubkey_valid(self) -> None:
        """Test if get_atomic_swap_batch_transaction allows one pubkey in case of only one direction of transfers."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_batch_transaction works with only to_pubkey
        tx = self.contract.get_atomic_swap_batch_transaction(
            self.ledger_api,
            contract_address="address",
            from_address="address",
            to_address="address",
            token_ids=[self.token_ids_a[0]],
            from_supplies=[0],
            to_supplies=[1],
            value=1,
            trade_nonce=0,
            to_pubkey="deadbeef",
        )
        assert tx is not None

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_cosmos_one_pubkey_invalid(self) -> None:
        """Test if get_atomic_swap_batch_transaction returns RuntimeError with missing from_pubkey."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_batch_transaction fails with missing from_key
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[1],
                to_supplies=[0],
                value=1,
                trade_nonce=0,
                to_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_ba_transaction_eth_no_signature(self) -> None:
        """Test if get_atomic_swap_single_transaction returns RuntimeError if signature not present on Ethereum case."""

        self.ledger_api.identifier = "ethereum"

        # Test if get_atomic_swap_single_transaction returns RuntimeError when signature is missing
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[1],
                to_supplies=[0],
                value=1,
                trade_nonce=0,
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_cosmos_to_pubkey_missing(self) -> None:
        """Test if get_atomic_swap_batch_transaction returns RuntimeError with missing to_pubkey."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with all amounts to be zero
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[1],
                to_supplies=[0],
                value=1,
                trade_nonce=0,
                from_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_cosmos_to_pubkey_missing_no_from_pubkey_required(
        self,
    ):
        """Test if get_atomic_swap_batch_transaction returns RuntimeError with missing to_pubkey and from_pubkey not required."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with missing to_pubkey
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[0],
                to_supplies=[1],
                value=1,
                trade_nonce=0,
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_cosmos_from_pubkey_missing_no_to_pubkey_required(
        self,
    ):
        """Test if get_atomic_swap_batch_transaction returns RuntimeError with missing from_pubkey and to_pubkey not required."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with missing from_pubkey
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[1],
                to_supplies=[0],
                value=0,
                trade_nonce=0,
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_cosmos_from_pubkey_only(self) -> None:
        """Test if get_atomic_swap_batch_transaction returns Tx in case with only from_pubkey."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction works with only from_pubkey
        res = self.contract.get_atomic_swap_batch_transaction(
            self.ledger_api,
            contract_address="address",
            from_address="address",
            to_address="address",
            token_ids=[self.token_ids_a[0]],
            from_supplies=[1],
            to_supplies=[0],
            value=0,
            trade_nonce=0,
            from_pubkey="deadbeef",
        )
        assert res is not None

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_amounts_missing(self) -> None:
        """Test if get_atomic_swap_single_transaction returns RuntimeError with missing amounts."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with all amounts to be zero
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=0,
                to_supply=0,
                value=0,
                trade_nonce=0,
                from_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_batch_transaction_amounts_missing(self) -> None:
        """Test if get_atomic_swap_batch_transaction returns RuntimeError with missing amounts."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with all amounts to be zero
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_batch_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_ids=[self.token_ids_a[0]],
                from_supplies=[0],
                to_supplies=[0],
                value=0,
                trade_nonce=0,
                from_pubkey="deadbeef",
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_cosmos_to_pubkey_missing_no_from_pubkey_required(
        self,
    ):
        """Test if get_atomic_swap_single_transaction returns RuntimeError with missing to_pubkey and from_pubkey not required."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with missing to_pubkey
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=0,
                to_supply=1,
                value=1,
                trade_nonce=0,
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_cosmos_from_pubkey_missing_no_to_pubkey_required(
        self,
    ):
        """Test if get_atomic_swap_single_transaction returns RuntimeError with missing from_pubkey and to_pubkey not required."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction fails with missing from_pubkey
        with pytest.raises(RuntimeError):
            self.contract.get_atomic_swap_single_transaction(
                self.ledger_api,
                contract_address="address",
                from_address="address",
                to_address="address",
                token_id=self.token_ids_a[0],
                from_supply=1,
                to_supply=0,
                value=0,
                trade_nonce=0,
            )

    @pytest.mark.ledger
    def test_get_atomic_swap_single_transaction_cosmos_from_pubkey_only(self) -> None:
        """Test if get_atomic_swap_single_transaction returns Tx in case with only from_pubkey."""

        self.ledger_api.identifier = "fetchai"

        # Test if get_atomic_swap_single_transaction works with only from_pubkey
        res = self.contract.get_atomic_swap_single_transaction(
            self.ledger_api,
            contract_address="address",
            from_address="address",
            to_address="address",
            token_id=self.token_ids_a[0],
            from_supply=1,
            to_supply=0,
            value=0,
            trade_nonce=0,
            from_pubkey="deadbeef",
        )
        assert res is not None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
Utils to support multiple CID versions.

Original implementation: https://github.com/ipld/py-cid/
"""

from abc import abstractmethod
from typing import Union, cast

import base58
import multibase
import multicodec
import multihash as mh
from morphys import ensure_bytes, ensure_unicode

DEFAULT_ENCODING = "base32"
BTC_ENCODING = "base58btc"


class BaseCID:
    """Base CID object."""

    _version: int
    _codec: str
    _multihash: bytes

    def __init__(self, version: int, codec: str, multihash: bytes):
        """Creates a new CID object."""

        self._version = version
        self._codec = codec
        self._multihash = ensure_bytes(multihash)

    @property
    def version(self) -> int:
        """CID version"""
        return self._version

    @property
    def codec(self) -> str:
        """CID codec"""
        return self._codec

    @property
    def multihash(self) -> bytes:
        """CID multihash"""
        return self._multihash

    @property
    @abstractmethod
    def buffer(self) -> bytes:
        """Multihash buffer."""

    @abstractmethod
    def encode(self, encoding: str = DEFAULT_ENCODING) -> bytes:
        """Encode multihash."""

    def __repr__(self) -> str:
        """Object representation."""

        def truncate(s: bytes, length: int) -> bytes:
            return s[:length] + b".." if len(s) > length else s

        truncate_length = 20
        multihash = (truncate(self._multihash, truncate_length),)

        return f"<version={self._version}, codec={self._codec}, multihash={multihash}>"

    def __str__(self) -> str:
        """String representation."""

        return ensure_unicode(self.encode())

    def __eq__(self, other: object) -> bool:
        """Dunder to check object equivalence."""

        if not isinstance(other, BaseCID):
            return NotImplemented

        return self.__dict__ == other.__dict__


class CIDv0(BaseCID):
    """CID version 0 object"""

    CODEC: str = "dag-pb"

    def __init__(self, multihash: bytes) -> None:
        """Initialize object."""

        super().__init__(0, self.CODEC, multihash)

    @property
    def buffer(self) -> bytes:
        """The raw representation that will be encoded."""

        return self.multihash

    def encode(self, encoding: str = DEFAULT_ENCODING) -> bytes:
        """base58-encoded buffer"""

        return ensure_bytes(base58.b58encode(self.buffer))

    def to_v1(self) -> "CIDv1":
        """Get an equivalent `CIDv1` object."""

        return CIDv1(self.CODEC, self.multihash)


class CIDv1(BaseCID):
    """CID version 1 object"""

    def __init__(self, codec: str, multihash: bytes) -> None:
        """Initialize object."""

        super().__init__(1, codec, multihash)

    @property
    def buffer(self) -> bytes:
        """The raw representation of the CID"""

        return b"".join(
            [bytes([self.version]), multicodec.add_prefix(self.codec, self.multihash)]
        )

    def encode(self, encoding: str = DEFAULT_ENCODING) -> bytes:
        """Encoded version of the raw representation"""

        return multibase.encode(encoding, self.buffer)

    def to_v0(self) -> CIDv0:
        """Get an equivalent `CIDv0` object."""

        if self.codec != CIDv0.CODEC:
            raise ValueError(
                "CIDv1 can only be converted for codec {}".format(CIDv0.CODEC)
            )

        return CIDv0(self.multihash)


CIDObject = Union[CIDv0, CIDv1]


class CID:
    """CID class."""

    @classmethod
    def make(
        cls,
        version: int,
        codec: str,
        multihash: bytes,
    ) -> CIDObject:
        """Make CID from given arguments."""

        if version not in (0, 1):
            raise ValueError(
                "version should be 0 or 1, {} was provided".format(version)
            )

        if not multicodec.is_codec(codec):
            raise ValueError("invalid codec {} provided, please check".format(codec))

        if not isinstance(multihash, bytes):
            raise ValueError("invalid type for multihash provided, should be bytes")

        if version == 1:
            return CIDv1(codec, multihash)

        if codec != CIDv0.CODEC:
            raise ValueError(
                "codec for version 0 can only be {}, found: {}".format(
                    CIDv0.CODEC, codec
                )
            )
        return CIDv0(multihash)

    @classmethod
    def is_cid(cls, cid: str) -> bool:
        """Checks if a given input string is valid encoded CID or not."""

        try:
            return bool(cls.from_string(cid))
        except ValueError:
            return False

    @classmethod
    def from_string(cls, cid: str) -> CIDObject:
        """Creates a CID object from a encoded form"""

        cid_bytes = ensure_bytes(cid, "utf-8")
        return cls.from_bytes(cid_bytes)

    @classmethod
    def from_bytes(cls, cid: bytes) -> CIDObject:
        """Creates a CID object from a encoded form"""

        if len(cid) < 2:
            raise ValueError("argument length can not be zero")

        # first byte for identity multibase and CIDv0 is 0x00
        # putting in assumption that multibase for CIDv0 can not be identity
        # refer: https://github.com/ipld/cid/issues/13#issuecomment-326490275

        if cid[0] != 0 and multibase.is_encoded(cid):
            # if the bytestream is multibase encoded
            cid = multibase.decode(cid)
            if len(cid) < 2:
                raise ValueError("cid length is invalid")

            data = cid[1:]
            version = int(cid[0])
            codec = multicodec.get_codec(data)
            multihash = multicodec.remove_prefix(data)

        elif cid[0] in (0, 1):
            # if the bytestream is a CID
            version = cid[0]
            data = cid[1:]
            codec = multicodec.get_codec(data)
            multihash = multicodec.remove_prefix(data)

        else:
            # otherwise its just base58-encoded multihash
            try:
                version = 0
                codec = CIDv0.CODEC
                multihash = base58.b58decode(cid)
            except ValueError:
                raise ValueError("multihash is not a valid base58 encoded multihash")

        mh.decode(multihash)
        return cls.make(version, codec, multihash)


def to_v0(hash_string: str) -> str:
    """Convert CID v1 hash to CID v0"""

    cid = CID.from_string(hash_string)
    if cid.version == 0:
        raise ValueError(f"{hash_string} is already v0.")

    return cast(CIDv1, cid).to_v0().encode().decode()


def to_v1(hash_string: str, encoding: str = DEFAULT_ENCODING) -> str:
    """Convert CID v0 hash to CID v1"""

    cid = CID.from_string(hash_string)
    if cid.version == 1:
        raise ValueError(f"{hash_string} is already v1.")

    return cast(CIDv0, cid).to_v1().encode(encoding=encoding).decode()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'aea scaffold' subcommand."""

import os
import re
import shutil
from datetime import datetime
from pathlib import Path
from typing import Optional, cast

import click
from jsonschema import ValidationError

from aea import AEA_DIR
from aea.cli.fingerprint import fingerprint_item
from aea.cli.packages import get_package_manager
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project, clean_after, pass_ctx
from aea.cli.utils.loggers import logger
from aea.cli.utils.package_utils import (
    create_symlink_packages_to_vendor,
    create_symlink_vendor_to_local,
    validate_package_name,
)
from aea.configurations.base import PackageType, PublicId
from aea.configurations.constants import (
    BUILD,
    CONNECTION,
    CONTRACT,
    CONTRACTS,
    CUSTOM,
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
    DEFAULT_VERSION,
    DOTTED_PATH_MODULE_ELEMENT_SEPARATOR,
    PACKAGE_TYPE_TO_CONFIG_FILE,
    PROTOCOL,
    SCAFFOLD_PUBLIC_ID,
    SKILL,
    _ETHEREUM_IDENTIFIER,
)
from aea.configurations.loader import ConfigLoader
from aea.helpers.io import open_file
from aea.helpers.ipfs.base import IPFSHashOnly

COPYRIGHT_HEADER = """\
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright {year} {author}
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""


def update_copyright_headers(ctx: Context, path: Path) -> None:
    """Update copyright headers"""

    # this allows for arbitrary authors
    year, author = datetime.now().year, ctx.agent_config.author
    top = "\n".join(COPYRIGHT_HEADER.split("\n")[:3])
    bottom = "\n".join(COPYRIGHT_HEADER.split("\n")[-15:])
    new_copyright_header = COPYRIGHT_HEADER.format(year=year, author=author)

    for file in path.rglob("**/*.py"):
        content = file.read_text()
        i, j = content.find(top), content.find(bottom)
        if i == j == -1:  # no copyright header present yet
            file.write_text(f"{new_copyright_header}\n{content}")
        elif i != j != -1:  # copyright header detected
            old_copyright_header = content[i : j + len(bottom)]
            content = content.replace(old_copyright_header, new_copyright_header)
            file.write_text(content)
        else:
            raise ValueError(f"Copyright pattern detection failed: {content}")


@click.group()
@click.option(
    "--to-local-registry",
    "-tlr",
    is_flag=True,
    help="Scaffold skill inside a local registry.",
)
@click.option(
    "--with-symlinks",
    is_flag=True,
    help="Add symlinks from vendor to non-vendor and packages to vendor folders.",
)
@click.pass_context
@check_aea_project
def scaffold(
    click_context: click.core.Context,
    with_symlinks: bool,
    to_local_registry: bool,
) -> None:  # pylint: disable=unused-argument
    """Scaffold a package for the agent."""
    ctx = cast(Context, click_context.obj)
    ctx.set_config("with_symlinks", with_symlinks)
    ctx.set_config("to_local_registry", to_local_registry)


@scaffold.command()
@click.argument("connection_name", type=str, required=True)
@pass_ctx
def connection(ctx: Context, connection_name: str) -> None:
    """Add a connection scaffolding to the configuration file and agent."""
    scaffold_item(ctx, CONNECTION, connection_name)


@scaffold.command()
@click.argument("contract_name", type=str, required=True)
@click.argument(
    "contract_abi_path", type=click.Path(dir_okay=True, exists=True), required=False
)
@pass_ctx
def contract(
    ctx: Context, contract_name: str, contract_abi_path: Optional[Path] = None
) -> None:
    """Add a contract scaffolding to the configuration file and agent."""
    scaffold_item(ctx, CONTRACT, contract_name)
    if contract_abi_path:
        add_contract_abi(ctx, contract_name, Path(contract_abi_path))


@scaffold.command()
@click.argument("name", type=str, required=True)
@pass_ctx
def custom(ctx: Context, name: str) -> None:
    """Scaffold a custom component package."""
    scaffold_item(ctx, CUSTOM, name)


@scaffold.command()
@click.argument("protocol_name", type=str, required=True)
@click.option("-y", "--yes", is_flag=True, default=False)
@pass_ctx
def protocol(ctx: Context, protocol_name: str, yes: bool) -> None:
    """Add a protocol scaffolding to the configuration file and agent."""
    if yes or click.confirm(
        "We highly recommend auto-generating protocols with the aea generate command. Do you really want to continue scaffolding?"
    ):
        scaffold_item(ctx, PROTOCOL, protocol_name)
    else:
        click.echo("Aborted. Exit")  # pragma: nocover


@scaffold.command()
@click.argument("skill_name", type=str, required=True)
@pass_ctx
def skill(ctx: Context, skill_name: str) -> None:
    """Add a skill scaffolding to the configuration file and agent."""
    scaffold_item(ctx, SKILL, skill_name)


@scaffold.command()
@pass_ctx
def decision_maker_handler(ctx: Context) -> None:
    """Add a decision maker scaffolding to the configuration file and agent."""
    _scaffold_dm_handler(ctx)


@scaffold.command()
@pass_ctx
def error_handler(ctx: Context) -> None:
    """Add an error scaffolding to the configuration file and agent."""
    _scaffold_error_handler(ctx)


@clean_after
def scaffold_item(ctx: Context, item_type: str, item_name: str) -> None:
    """
    Add an item scaffolding to the configuration file and agent.

    :param ctx: Context object.
    :param item_type: type of item.
    :param item_name: item name.

    :raises ClickException: if some error occurs.
    """
    validate_package_name(item_name)
    author_name = ctx.agent_config.author
    loader = getattr(ctx, f"{item_type}_loader")
    default_config_filename = PACKAGE_TYPE_TO_CONFIG_FILE[item_type]

    to_local_registry = ctx.config.get("to_local_registry")

    item_type_plural = item_type + "s"
    existing_ids = getattr(ctx.agent_config, f"{item_type}s")
    existing_ids_only_author_and_name = map(lambda x: (x.author, x.name), existing_ids)
    # check if we already have an item with the same public id
    if (author_name, item_name) in existing_ids_only_author_and_name:
        raise click.ClickException(
            f"A {item_type} with name '{item_name}' already exists. Aborting..."
        )

    # create the item folder
    if to_local_registry:
        click.echo(f"Adding {item_type} scaffold '{item_name}' to local registry...")
        registry_path = Path(ctx.registry_path)
        dest = registry_path / ctx.agent_config.author / item_type_plural
    else:
        click.echo(
            f"Adding {item_type} scaffold '{item_name}' to the agent '{ctx.agent_config.agent_name}'..."
        )
        dest = Path(item_type_plural)

    dest.mkdir(exist_ok=True, parents=True)
    dest = dest / item_name

    if os.path.exists(dest):
        raise click.ClickException(
            f"A {item_type} with this name already exists. Please choose a different name and try again."
        )

    ctx.clean_paths.append(str(dest))
    try:
        # copy the item package into the agent project.
        src = Path(os.path.join(AEA_DIR, item_type_plural, "scaffold"))
        logger.debug(f"Copying {item_type} modules. src={src} dst={dest}")
        shutil.copytree(src, dest)
        update_copyright_headers(ctx, dest)

        new_public_id = PublicId(author_name, item_name, DEFAULT_VERSION)

        # ensure the name in the yaml and the name of the folder are the same
        if to_local_registry:
            config_filepath = dest / default_config_filename
        else:
            config_filepath = Path(
                ctx.cwd, item_type_plural, item_name, default_config_filename
            )

        with open_file(config_filepath) as fp:
            config = loader.load(fp)
        config.name = item_name
        config.author = author_name
        with open_file(config_filepath, "w") as fp:
            loader.dump(config, fp)

        # update 'PUBLIC_ID' variable with the right public id in connection.py!

        for file_name in ["__init__.py", "connection.py"]:
            file_path = Path(dest) / file_name
            if not file_path.exists():
                continue
            py_file = Path(file_path)
            py_file.write_text(
                re.sub(SCAFFOLD_PUBLIC_ID, str(new_public_id), py_file.read_text())
            )

        # fingerprint item.
        if to_local_registry:
            ctx.cwd = str(registry_path / author_name)
            fingerprint_item(ctx, item_type, new_public_id)
        else:
            fingerprint_item(ctx, item_type, new_public_id)

        package_hash = IPFSHashOnly().hash_directory(str(dest))
        new_public_id_with_hash = PublicId(
            author_name, item_name, DEFAULT_VERSION, package_hash
        )
        if to_local_registry:
            get_package_manager(package_dir=registry_path).register(
                dest,
                package_type=PackageType(item_type),
            ).dump()
        else:
            logger.debug(f"Registering the {item_type} into {DEFAULT_AEA_CONFIG_FILE}")
            existing_ids.add(new_public_id_with_hash)
            with open_file(os.path.join(ctx.cwd, DEFAULT_AEA_CONFIG_FILE), "w") as fp:
                ctx.agent_loader.dump(ctx.agent_config, fp)

        if ctx.config.get("with_symlinks", False):
            click.echo(
                "Adding symlinks from vendor to non-vendor and packages to vendor folders."
            )
            create_symlink_vendor_to_local(ctx, item_type, new_public_id)
            create_symlink_packages_to_vendor(ctx)

    except ValidationError:
        raise click.ClickException(
            f"Error when validating the {item_type} configuration file."
        )
    except Exception as e:
        raise click.ClickException(str(e))


def _scaffold_dm_handler(ctx: Context) -> None:
    """Scaffold the decision maker handler."""
    _scaffold_non_package_item(
        ctx,
        "decision_maker_handler",
        "decision maker handler",
        "DecisionMakerHandler",
        "decision_maker",
    )


def _scaffold_error_handler(ctx: Context) -> None:
    """Scaffold the error handler."""
    _scaffold_non_package_item(
        ctx, "error_handler", "error handler", "ErrorHandler", "error_handler"
    )


def _scaffold_non_package_item(
    ctx: Context, item_type: str, type_name: str, class_name: str, aea_dir: str
) -> None:
    """
    Scaffold a non-package item (e.g. decision maker handler, or error handler).

    :param ctx: the CLI context.
    :param item_type: the item type (e.g. 'decision_maker_handler')
    :param type_name: the type name (e.g. "decision maker")
    :param class_name: the class name (e.g. "DecisionMakerHandler")
    :param aea_dir: the AEA directory that contains the scaffold module
    """
    existing_item = getattr(ctx.agent_config, item_type)
    if existing_item != {}:
        raise click.ClickException(
            f"A {type_name} specification already exists. Aborting..."
        )

    dest = Path(f"{item_type}.py")
    agent_name = ctx.agent_config.agent_name
    click.echo(f"Adding {type_name} scaffold to the agent '{agent_name}'...")
    # create the file name
    dotted_path = f".{item_type}{DOTTED_PATH_MODULE_ELEMENT_SEPARATOR}{class_name}"
    try:
        # copy the item package into the agent project.
        src = Path(os.path.join(AEA_DIR, aea_dir, "scaffold.py"))
        logger.debug(f"Copying {type_name}. src={src} dst={dest}")
        shutil.copyfile(src, dest)

        # add the item to the configurations.
        logger.debug(f"Registering the {type_name} into {DEFAULT_AEA_CONFIG_FILE}")
        setattr(
            ctx.agent_config,
            item_type,
            {
                "dotted_path": str(dotted_path),
                "file_path": str(os.path.join(".", dest)),
                "config": {},
            },
        )
        ctx.agent_loader.dump(
            ctx.agent_config,
            open_file(os.path.join(ctx.cwd, DEFAULT_AEA_CONFIG_FILE), "w"),
        )

    except Exception as e:
        os.remove(dest)
        raise click.ClickException(str(e))


def add_contract_abi(ctx: Context, contract_name: str, contract_abi_path: Path) -> None:
    """
    Add the contract ABI to a contract scaffold.

    :param ctx: the CLI context.
    :param contract_name: the contract name.
    :param contract_abi_path: the contract ABI path.
    """

    # Get some data from the context
    to_local_registry = ctx.config.get("to_local_registry")
    author_name = ctx.agent_config.author

    # Create the build directory and copy the ABI file
    contract_dir_root = (
        Path(str(ctx.agent_config.directory)) if to_local_registry else Path()
    )

    contract_dir = contract_dir_root / Path(CONTRACTS) / Path(contract_name)
    abi_dest = contract_dir / Path(BUILD) / Path(contract_abi_path.name)

    click.echo(f"Updating contract scaffold '{contract_name}' to include ABI...")

    os.makedirs(os.path.dirname(abi_dest), exist_ok=True)
    shutil.copyfile(str(contract_abi_path), abi_dest)

    # Load configuration (contract.yaml)
    config_file = Path(contract_dir / DEFAULT_CONTRACT_CONFIG_FILE)
    config_loader = ConfigLoader.from_configuration_type(PackageType.CONTRACT)
    contract_config = config_loader.load(config_file.open("r"))

    # Add ABI to the configuration
    # Can't use contract_config.update() here. Since contract_interface_paths is empty during instantiation,
    # it cannot be validated afterwards even if we add the correct values to ContractConfig.FIELDS_ALLOWED_TO_UPDATE
    contract_config.contract_interface_paths = {
        _ETHEREUM_IDENTIFIER: f"{BUILD}/{contract_abi_path.name}"
    }

    # Write the new configuration
    config_loader.dump(contract_config, config_file.open("w+"))

    # Fingerprint again
    new_public_id = PublicId(author_name, contract_name, DEFAULT_VERSION)

    if to_local_registry:
        ctx.cwd = str(ctx.agent_config.directory)
        get_package_manager(
            package_dir=cast(Path, ctx.agent_config.directory).parent
        ).register(
            contract_dir,
            package_type=PackageType.CONTRACT,
        ).dump()

    fingerprint_item(ctx, CONTRACT, new_public_id)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the `aea launch` sub-command."""

# import json
# import os
# import sqlite3
# import sys
# import uuid

# import pytest
# from aea_ledger_fetchai import FetchAICrypto

# from aea.test_tools.test_cases import AEATestCaseMany

# from tests.common.pexpect_popen import PexpectWrapper
# from tests.conftest import FETCHAI_PRIVATE_KEY_FILE_CONNECTION


# class TestLaunchEndToEnd(AEATestCaseMany):
#     """Perform aea launch end to end test."""

#     key = "seller_service"
#     value = None

#     registration_agent_connection = {
#         "delegate_uri": "127.0.0.1:11011",
#         "entry_peers": [],
#         "ledger_id": "fetchai",
#         "local_uri": "127.0.0.1:9011",
#         "log_file": "libp2p_node.log",
#         "public_uri": "127.0.0.1:9011",
#     }

#     search_agent_connection = {
#         "delegate_uri": "127.0.0.1:11012",
#         "entry_peers": [],
#         "ledger_id": "fetchai",
#         "local_uri": "127.0.0.1:9012",
#         "log_file": "libp2p_node.log",
#         "public_uri": "127.0.0.1:9012",
#     }

#     @pytest.mark.integration
#     def test_end_to_end(self):
#         """Perform end to end test with simple register/search agents."""
#         registration_agent_name = "registration_agent"
#         self.value = uuid.uuid4().hex
#         self.fetch_agent(
#             "fetchai/simple_service_registration",
#             agent_name=registration_agent_name,
#             is_local=True,
#         )

#         self.run_cli_command(
#             "config",
#             "set",
#             "vendor.fetchai.connections.p2p_libp2p.config",
#             "--type",
#             "dict",
#             json.dumps(self.registration_agent_connection),
#             cwd=registration_agent_name,
#         )
#         self.run_cli_command(
#             "config",
#             "set",
#             "vendor.fetchai.skills.simple_service_registration.models.strategy.args.service_data",
#             "--type",
#             "dict",
#             json.dumps({"key": self.key, "value": self.value}),
#             cwd=registration_agent_name,
#         )
#         self.run_cli_command(
#             "config",
#             "set",
#             "vendor.fetchai.connections.soef.config.token_storage_path",
#             os.path.join(self.t, registration_agent_name, "soef_key.txt"),
#             cwd=registration_agent_name,
#         )

#         storage_file_name = os.path.abspath(
#             os.path.join(registration_agent_name, "test.db")
#         )
#         self.run_cli_command(
#             "config",
#             "set",
#             "agent.storage_uri",
#             f"sqlite://{storage_file_name}",
#             cwd=registration_agent_name,
#         )

#         search_agent_name = "search_agent"
#         self.fetch_agent(
#             "fetchai/simple_service_search",
#             agent_name=search_agent_name,
#             is_local=True,
#         )
#         self.run_cli_command(
#             "config",
#             "set",
#             "vendor.fetchai.connections.p2p_libp2p.config",
#             "--type",
#             "dict",
#             json.dumps(self.search_agent_connection),
#             cwd=search_agent_name,
#         )
#         self.run_cli_command(
#             "config",
#             "set",
#             "vendor.fetchai.skills.simple_service_search.models.strategy.args.search_query",
#             "--type",
#             "dict",
#             json.dumps(
#                 {
#                     "constraint_type": "==",
#                     "search_key": self.key,
#                     "search_value": self.value,
#                 }
#             ),
#             cwd=search_agent_name,
#         )
#         self.run_cli_command(
#             "config",
#             "set",
#             "vendor.fetchai.skills.simple_service_search.behaviours.service_search.args.tick_interval",
#             "--type",
#             "int",
#             "2",
#             cwd=search_agent_name,
#         )
#         self.run_cli_command(
#             "config",
#             "set",
#             "vendor.fetchai.connections.soef.config.token_storage_path",
#             os.path.join(self.t, search_agent_name, "soef_key.txt"),
#             cwd=search_agent_name,
#         )
#         self.run_cli_command(
#             "build", cwd=registration_agent_name,
#         )
#         self.run_cli_command(
#             "build", cwd=search_agent_name,
#         )
#         self.set_agent_context(registration_agent_name)
#         self.generate_private_key(
#             FetchAICrypto.identifier, FETCHAI_PRIVATE_KEY_FILE_CONNECTION
#         )
#         self.add_private_key(
#             FetchAICrypto.identifier,
#             FETCHAI_PRIVATE_KEY_FILE_CONNECTION,
#             connection=True,
#         )
#         self.generate_private_key()
#         self.add_private_key()
#         self.unset_agent_context()
#         self.run_cli_command(
#             "issue-certificates", cwd=registration_agent_name,
#         )
#         self.set_agent_context(search_agent_name)
#         self.generate_private_key(
#             FetchAICrypto.identifier, FETCHAI_PRIVATE_KEY_FILE_CONNECTION
#         )
#         self.add_private_key(
#             FetchAICrypto.identifier,
#             FETCHAI_PRIVATE_KEY_FILE_CONNECTION,
#             connection=True,
#         )
#         self.generate_private_key()
#         self.add_private_key()
#         self.unset_agent_context()
#         self.run_cli_command(
#             "issue-certificates", cwd=search_agent_name,
#         )

#         proc = PexpectWrapper(  # nosec
#             [
#                 sys.executable,
#                 "-m",
#                 "aea.cli",
#                 "-v",
#                 "DEBUG",
#                 "launch",
#                 registration_agent_name,
#                 search_agent_name,
#             ],
#             env=os.environ,
#             maxread=10000,
#             encoding="utf-8",
#             logfile=sys.stdout,
#         )
#         try:
#             proc.expect_all(
#                 [f"[{search_agent_name}] found number of agents=1, search_response"],
#                 timeout=30,
#             )
#         finally:
#             proc.control_c()
#             proc.expect("Exit cli. code: 0", timeout=30)

#         assert os.path.exists(storage_file_name)
#         con = sqlite3.connect(storage_file_name)
#         try:
#             cursor = con.cursor()
#             tables = cursor.execute(
#                 "SELECT name FROM sqlite_master WHERE type='table';"
#             ).fetchall()
#             assert tables
#             table_name = tables[0][0]
#             num_of_records = cursor.execute(  # nosec
#                 f"SELECT count(*) FROM {table_name};"
#             ).fetchone()[0]
#             assert num_of_records > 0
#         finally:
#             con.close



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for config data types."""

import copy
import operator
from itertools import permutations
from math import isnan
from typing import Tuple

import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

from aea.configurations.data_types import (
    Dependency,
    PackageId,
    PackageVersion,
    PublicId,
)

from tests.conftest import MAX_FLAKY_RERUNS
from tests.strategies.data_types import (
    package_id_strategy,
    public_id_strategy,
    version_info_strategy,
)

NUMERIC_TYPES = int, float
SEQUENCE_TYPES = str, list, tuple
ITERATOR_TYPES = *SEQUENCE_TYPES, set, frozenset, zip, map
CUSTOM_TYPES = PublicId, PackageId, Dependency
ALL_TYPES = *NUMERIC_TYPES, *ITERATOR_TYPES, *CUSTOM_TYPES
COMPARISON_OPERATORS = [
    operator.lt,
    operator.le,
    operator.eq,
    operator.ne,
    operator.ge,
    operator.gt,
]


def all_comparisons_operations_equal(pair_a, pair_b) -> bool:
    """Check whether all operator comparisons return the same result."""

    def all_comparisons(pair) -> Tuple[bool, ...]:
        return tuple(f(*pair) for f in COMPARISON_OPERATORS)

    return all_comparisons(pair_a) == all_comparisons(pair_b)


@pytest.mark.parametrize("self_type", [PackageVersion, PublicId, PackageId])
def test_self_type_comparison(self_type):
    """Test comparison to self"""

    self = st.from_type(self_type).example()
    other = copy.deepcopy(self)
    for f in COMPARISON_OPERATORS:
        assert isinstance(f(self, other), bool)
        assert isinstance(f(other, self), bool)


# possible issue with bad generated version specifier
@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@given(
    st.one_of(list(map(st.from_type, ALL_TYPES))),
    st.one_of(list(map(st.from_type, CUSTOM_TYPES))),
)
def test_type_comparison(self, other):
    """Test type comparison"""

    if isinstance(self, float):
        assume(not isnan(self))
    assume(type(self) is not type(other))

    funcs = (operator.le, operator.lt, operator.ge, operator.gt)

    assert not isinstance(self, type(other))
    assert not isinstance(other, type(self))
    assert self.__eq__(self)
    assert other.__eq__(other)
    assert self.__ne__(other)
    assert other.__ne__(self)

    for f in funcs:
        with pytest.raises((TypeError, ValueError)):
            assert f(self, other)


@given(st.tuples(version_info_strategy, version_info_strategy))
def test_package_version_comparison(version_pair):
    """Test package version comparison"""

    package_pair = tuple(map(PackageVersion, version_pair))
    assert all_comparisons_operations_equal(version_pair, package_pair)


@given(st.tuples(public_id_strategy, public_id_strategy))
def test_public_id_comparison(public_id_pair):
    """Test public id comparison"""

    public_id_pair[0]._name = public_id_pair[1]._name
    public_id_pair[0]._author = public_id_pair[1]._author
    version_pair = tuple(p.package_version._version for p in public_id_pair)
    assert all_comparisons_operations_equal(version_pair, public_id_pair)


@given(st.tuples(package_id_strategy, package_id_strategy))
def test_package_id_comparison(package_id_pair):
    """Test package id comparison"""

    package_id_pair[0]._public_id = package_id_pair[1]._public_id
    package_id_pair[0]._package_type = package_id_pair[1]._package_type
    version_pair = tuple(p.public_id.package_version._version for p in package_id_pair)
    assert all_comparisons_operations_equal(version_pair, package_id_pair)


@pytest.mark.parametrize("version_like", ["any", "latest", "0.1.0"])
def test_any_and_latest_equal(version_like: str):
    """Test special version types "any" and "latest" when equal."""

    version_a, version_b = (PackageVersion(version_like) for _ in range(2))
    assert version_a == version_b
    assert not version_a < version_b


@pytest.mark.parametrize(
    "version_like_pair", permutations(["any", "latest", "0.1.0"], 2)
)
def test_any_latest_and_numeric_unequal(version_like_pair: Tuple[str]):
    """Test special version types "any" and "latest" when equal."""

    self, other = map(PackageVersion, version_like_pair)
    assert not self == other
    assert self != other

    funcs = (operator.le, operator.lt, operator.ge, operator.gt)
    for f in funcs:
        with pytest.raises(TypeError, match="not supported between"):
            assert f(self, other)


class TestDependency:
    """Test Dependency class."""

    def test_parse_from_string(self) -> None:
        """Test from_string method."""

        string = "tomte==0.6.1"
        dep = Dependency.from_string(string=string)
        assert dep.name == "tomte"
        assert str(dep.version) == "==0.6.1"
        assert dep.to_pip_string() == string

    def test_parse_from_string_wth_extras(self) -> None:
        """Test from_string method."""

        string = "tomte[tox,tests]==0.6.1"
        dep = Dependency.from_string(string=string)
        assert dep.name == "tomte"
        assert str(dep.version) == "==0.6.1"
        assert dep.extras == ["tox", "tests"]
        assert dep.to_pip_string() == string

    def test_parse_from_git_string(self) -> None:
        """Test from_string method."""

        string = "git+https://github.com/valory-xyz/tomte.git@03adee2a0a5681c6d8e77160be27edffaa8e3bee#egg=tomte"
        dep = Dependency.from_string(string=string)
        assert dep.name == "tomte"
        assert str(dep.ref) == "03adee2a0a5681c6d8e77160be27edffaa8e3bee"
        assert dep.git == "https://github.com/valory-xyz/tomte.git"
        assert dep.to_pip_string() == string

    def test_parse_from_pipfile_specifier(self) -> None:
        """Test from_pipfile_specifier method"""

        string = 'tomte = "==0.6.1"'
        dep = Dependency.from_pipfile_string(string)
        assert dep.name == "tomte"
        assert str(dep.version) == "==0.6.1"
        assert dep.to_pipfile_string() == string

    def test_parse_from_pipfile_specifier_with_extras(self) -> None:
        """Test from_pipfile_specifier method"""

        string = 'tomte = {version = "==0.6.1", extras = ["tox", "tests"]}'
        dep = Dependency.from_pipfile_string(string)
        assert dep.name == "tomte"
        assert str(dep.version) == "==0.6.1"
        assert dep.extras == ["tox", "tests"]
        assert dep.to_pipfile_string() == string

    def test_parse_from_pipfile_specifier_with_git_ref(self) -> None:
        """Test from_pipfile_specifier method"""

        string = 'tomte = {ref = "03adee2a0a5681c6d8e77160be27edffaa8e3bee", git = "git+https://github.com/valory-xyz/tomte.git"}'
        dep = Dependency.from_pipfile_string(string)
        assert dep.name == "tomte"
        assert str(dep.ref) == "03adee2a0a5681c6d8e77160be27edffaa8e3bee"
        assert dep.git == "https://github.com/valory-xyz/tomte.git"
        assert dep.to_pipfile_string() == string



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains tests for Libp2p tcp client connection."""

# pylint: skip-file

import time
from unittest import mock
from unittest.mock import call

import pytest

from aea.configurations.constants import DEFAULT_LEDGER
from aea.mail.base import Empty
from aea.test_tools.mocks import RegexComparator

from packages.valory.connections.p2p_libp2p.tests.base import libp2p_log_on_failure_all
from packages.valory.connections.test_libp2p.tests.base import (
    BaseP2PLibp2pTest,
    LIBP2P_LEDGER,
    TIMEOUT,
    _make_libp2p_client_connection,
    _make_libp2p_connection,
    ports,
)
from packages.valory.connections.test_libp2p.tests.test_p2p_libp2p.test_communication import (
    TestP2PLibp2pConnectionRouting,
)

DEFAULT_CLIENTS_PER_NODE = 1


@pytest.mark.asyncio
@libp2p_log_on_failure_all
class TestLibp2pClientConnectionConnectDisconnect(BaseP2PLibp2pTest):
    """Test that connection is established and torn down correctly"""

    @pytest.mark.asyncio
    async def test_libp2p_client_connection_connect_disconnect(self):
        """Test connnect then disconnect."""

        delegate_port = next(ports)
        connection_node = _make_libp2p_connection(
            delegate=True, delegate_port=delegate_port
        )
        connection = _make_libp2p_client_connection(
            peer_public_key=connection_node.node.pub,
            node_port=delegate_port,
        )

        assert connection.is_connected is False
        try:
            await connection_node.connect()
            await connection.connect()
            assert connection.is_connected is True
            await connection.disconnect()
            assert connection.is_connected is False
        except Exception:
            raise
        finally:
            await connection_node.disconnect()


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionEchoEnvelope(BaseP2PLibp2pTest):
    """Test that connection will route envelope to destination through the same libp2p node"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        super().setup_class()

        delegate_port = next(ports)
        cls.connection_node = cls.make_connection(
            delegate=True,
            delegate_port=delegate_port,
        )

        cls.connection_client_1 = cls.make_client_connection(
            peer_public_key=cls.connection_node.node.pub,
            ledger_api_id=cls.cosmos_crypto.identifier,
            node_port=delegate_port,
        )
        cls.connection_client_2 = cls.make_client_connection(
            peer_public_key=cls.connection_node.node.pub,
            ledger_api_id=cls.ethereum_crypto.identifier,
            node_port=delegate_port,
        )

    def test_connection_is_established(self):
        """Test connection is established."""
        assert self.all_connected

    def test_envelope_routed(self):
        """Test the envelope is routed."""

        sender = self.connection_client_1.address
        to = self.connection_client_2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[1].put(envelope)
        delivered_envelope = self.multiplexers[2].get(block=True, timeout=TIMEOUT)
        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

    def test_envelope_echoed_back(self):
        """Test the envelope is echoed back."""

        sender = self.connection_client_1.address
        to = self.connection_client_2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[1].put(envelope)
        delivered_envelope = self.multiplexers[2].get(block=True, timeout=TIMEOUT)
        assert delivered_envelope is not None

        delivered_envelope.to = sender
        delivered_envelope.sender = to

        self.multiplexers[2].put(delivered_envelope)
        echoed_envelope = self.multiplexers[1].get(block=True, timeout=TIMEOUT)
        assert self.sent_is_echoed_envelope(envelope, echoed_envelope)

    def test_envelope_echoed_back_node_agent(self):
        """Test the envelope is echoed back."""

        sender = self.connection_client_1.address
        to = self.connection_node.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[1].put(envelope)
        delivered_envelope = self.multiplexers[0].get(block=True, timeout=TIMEOUT)
        assert delivered_envelope is not None

        delivered_envelope.to = sender
        delivered_envelope.sender = to

        self.multiplexers[0].put(delivered_envelope)
        echoed_envelope = self.multiplexers[1].get(block=True, timeout=TIMEOUT)
        assert self.sent_is_echoed_envelope(envelope, echoed_envelope)


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionEchoEnvelopeTwoDHTNode(BaseP2PLibp2pTest):
    """Test that connection will route envelope to destination connected to different node"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        super().setup_class()

        cls.ports = next(ports), next(ports)

        cls.connection_node_1 = cls.make_connection(
            delegate_port=cls.ports[0],
            delegate=True,
        )
        genesis_peer = cls.connection_node_1.node.multiaddrs[0]
        cls.connection_node_2 = cls.make_connection(
            delegate_port=cls.ports[1],
            entry_peers=[genesis_peer],
            delegate=True,
        )

        cls.connection_client_1 = cls.make_client_connection(
            peer_public_key=cls.connection_node_1.node.pub,
            node_port=cls.ports[0],
        )
        cls.connection_client_2 = cls.make_client_connection(
            peer_public_key=cls.connection_node_2.node.pub,
            node_port=cls.ports[1],
        )

    def test_connection_is_established(self):
        """Test the connection is established."""
        assert self.all_connected

    def test_envelope_routed(self):
        """Test the envelope is routed."""

        sender = self.connection_client_1.address
        to = self.connection_client_2.address

        envelope = self.enveloped_default_message(to=to, sender=sender)
        self.multiplexers[2].put(envelope)
        delivered_envelope = self.multiplexers[3].get(block=True, timeout=20)

        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

    def test_envelope_echoed_back(self):
        """Test the envelope is echoed back."""

        sender = self.connection_client_1.address
        to = self.connection_client_2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[2].put(envelope)
        delivered_envelope = self.multiplexers[3].get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = sender
        delivered_envelope.sender = to

        self.multiplexers[3].put(delivered_envelope)
        echoed_envelope = self.multiplexers[2].get(block=True, timeout=5)
        assert self.sent_is_echoed_envelope(envelope, echoed_envelope)

    def test_envelope_echoed_back_node_agent(self):
        """Test the envelope is echoed back node agent."""

        sender = self.connection_client_1.address
        to = self.connection_node_2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[2].put(envelope)
        delivered_envelope = self.multiplexers[1].get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = sender
        delivered_envelope.sender = to

        self.multiplexers[1].put(delivered_envelope)
        echoed_envelope = self.multiplexers[2].get(block=True, timeout=5)

        assert self.sent_is_echoed_envelope(envelope, echoed_envelope)


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionRouting(TestP2PLibp2pConnectionRouting):
    """Test that libp2p DHT network will reliably route envelopes from clients connected to different nodes"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        BaseP2PLibp2pTest.setup_class()

        cls.ports = next(ports), next(ports)
        connection1 = cls.make_connection(
            delegate_port=cls.ports[0],
            delegate=True,
        )
        entry_peer = connection1.node.multiaddrs[0]
        connection2 = cls.make_connection(
            delegate_port=cls.ports[1],
            entry_peers=[entry_peer],
            delegate=True,
        )

        peers_public_keys = [connection1.node.pub, connection2.node.pub]
        assert len(cls.ports) == len(peers_public_keys)
        for port, peer_public_key in zip(cls.ports, peers_public_keys):
            cls.make_client_connection(
                peer_public_key=peer_public_key,
                node_port=port,
            )


@libp2p_log_on_failure_all
class BaseTestLibp2pClientSamePeer(BaseP2PLibp2pTest):
    """Base test class for reconnection tests."""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        super().setup_class()

        cls.delegate_port = next(ports)
        cls.connection_node = cls.make_connection(
            delegate=True,
            delegate_port=cls.delegate_port,
        )

        cls.connection_client_1 = cls.make_client_connection(
            peer_public_key=cls.connection_node.node.pub,
            ledger_api_id=LIBP2P_LEDGER,
            node_port=cls.delegate_port,
        )
        cls.connection_client_2 = cls.make_client_connection(
            peer_public_key=cls.connection_node.node.pub,
            ledger_api_id=DEFAULT_LEDGER,
            node_port=cls.delegate_port,
        )


@libp2p_log_on_failure_all
class TestLibp2pClientReconnectionSendEnvelope(BaseTestLibp2pClientSamePeer):
    """Test that connection will send envelope with error, and that reconnection fixes it."""

    def test_envelope_sent(self):
        """Test the envelope is routed."""
        sender = self.connection_client_1.address
        to = self.connection_client_2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        # cause failure on send
        with mock.patch.object(
            self.connection_client_1.logger, "exception"
        ) as _mock_logger, mock.patch.object(
            self.connection_client_1._node_client, "_write", side_effect=Exception
        ):
            self.multiplexers[1].put(envelope)
            delivered_envelope = self.multiplexers[2].get(block=True, timeout=TIMEOUT)
            expected = "Exception raised on message send. Try reconnect and send again."
            _mock_logger.assert_has_calls([call(expected)])

        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)


@libp2p_log_on_failure_all
class TestLibp2pClientReconnectionReceiveEnvelope(BaseTestLibp2pClientSamePeer):
    """Test that connection will receive envelope with error, and that reconnection fixes it."""

    def test_envelope_received(self):
        """Test the envelope is routed."""
        sender = self.connection_client_2.address
        to = self.connection_client_1.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        # make the reception fail
        with mock.patch.object(
            self.connection_client_1.logger, "error"
        ) as _mock_logger, mock.patch.object(
            self.connection_client_1._node_client,
            "_read",
            side_effect=ConnectionError(),
        ):
            # this envelope will be lost.
            self.multiplexers[2].put(envelope)
            time.sleep(2.0)
            expected = "Connection error:.*Try to reconnect and read again"
            _mock_logger.assert_has_calls([call(RegexComparator(expected))])

        # proceed as usual. Now we expect the connection to have reconnected successfully
        delivered_envelope = self.multiplexers[1].get(block=True, timeout=20)
        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)


@libp2p_log_on_failure_all
class TestLibp2pClientEnvelopeOrderSamePeer(BaseTestLibp2pClientSamePeer):
    """Test that the order of envelope is the guaranteed to be the same."""

    NB_ENVELOPES = 1000

    def test_burst_order(self):
        """Test order of envelope burst is guaranteed on receiving end."""

        sender = self.connection_client_1.address
        to = self.connection_client_2.address
        sent_envelopes = [
            self.enveloped_default_message(to, sender) for _ in range(self.NB_ENVELOPES)
        ]

        for envelope in sent_envelopes:
            envelope.message = envelope.message_bytes  # encode
            self.multiplexers[1].put(envelope)

        received_envelopes = []
        for _ in range(self.NB_ENVELOPES):
            envelope = self.multiplexers[2].get(block=True, timeout=3)
            received_envelopes.append(envelope)

        # test no new message is "created"
        with pytest.raises(Empty):
            self.multiplexers[2].get(block=True, timeout=1)

        assert len(sent_envelopes) == len(received_envelopes)
        for expected, actual in zip(sent_envelopes, received_envelopes):
            assert expected.message == actual.message



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `aea packages sync` command."""

import json
import logging
import re
from unittest import mock

import click
import pytest

from aea.configurations.constants import PACKAGES
from aea.configurations.data_types import PackageId
from aea.package_manager.base import PACKAGES_FILE
from aea.package_manager.v1 import PackageManagerV1
from aea.test_tools.test_cases import BaseAEATestCase


@mock.patch("aea.package_manager.base.load_fetch_ipfs")
@mock.patch(
    "aea.package_manager.base.BasePackageManager.calculate_hash_from_package_id"
)
class TestSyncCommand(BaseAEATestCase):
    """Test sync command."""

    def test_sync_normal(self, ipfs_mock, hash_mock, caplog) -> None:
        """Test sync command."""

        with caplog.at_level(logging.INFO):
            result = self.run_cli_command("packages", "sync")
            assert result.exit_code == 0
            assert "No package was updated." in caplog.text

    def test_sync_with_missing_dev_packages(self, ipfs_mock, hash_mock, caplog) -> None:
        """Test sync with missing packages."""

        packages_file = self.t / PACKAGES / PACKAGES_FILE
        packages = dict(
            map(
                lambda x: (PackageId.from_uri_path(x[0]), x[1]),
                json.loads(packages_file.read_text())["dev"].items(),
            )
        )
        packages[PackageId.from_uri_path("skill/author/some_skill/0.1.0")] = (
            "bafybeidv77u2xl52mnxakwvh7fuh46aiwfpteyof4eaptfd4agoi6cdblb"
        )

        with mock.patch.object(
            PackageManagerV1, "dev_packages", new=packages
        ), caplog.at_level(logging.INFO), mock.patch.object(
            PackageManagerV1, "add_package"
        ):
            result = self.run_cli_command("packages", "sync", "--dev")
            assert result.exit_code == 0
            assert (
                "(skill, author/some_skill:0.1.0) not found locally, downloading..."
                in caplog.text
            )

    def test_sync_with_missing_third_party_packages(
        self, ipfs_mock, hash_mock, caplog
    ) -> None:
        """Test sync with missing packages."""

        packages = {
            PackageId.from_uri_path(
                "skill/author/some_skill/0.1.0"
            ): "bafybeidv77u2xl52mnxakwvh7fuh46aiwfpteyof4eaptfd4agoi6cdblb"
        }

        with mock.patch.object(
            PackageManagerV1, "third_party_packages", new=packages
        ), caplog.at_level(logging.INFO), mock.patch.object(
            PackageManagerV1, "add_package"
        ):
            result = self.run_cli_command("packages", "sync")
            assert result.exit_code == 0
            assert (
                "(skill, author/some_skill:0.1.0) not found locally, downloading..."
                in caplog.text
            )

    def test_sync_with_wrong_hash(self, ipfs_mock, hash_mock, caplog) -> None:
        """Test sync with missing packages."""

        packages = {
            PackageId.from_uri_path(
                "protocol/open_aea/signing/1.0.0"
            ): "bafybeiambqptflge33eemdhis2whik67hjplfnqwieoa6wblzlaf7vuo41"
        }

        with mock.patch.object(PackageManagerV1, "third_party_packages", new=packages):
            with pytest.raises(
                click.ClickException,
                match=re.escape(
                    "Hashes for (protocol, open_aea/signing:1.0.0) does not match;"
                ),
            ):
                self.run_cli_command("packages", "sync")

        with mock.patch.object(
            PackageManagerV1, "third_party_packages", new=packages
        ), mock.patch("shutil.rmtree"), caplog.at_level(logging.INFO):
            result = self.run_cli_command(
                "packages",
                "sync",
                "--update-packages",
            )
            assert result.exit_code == 0
            assert (
                "Hash does not match for (protocol, open_aea/signing:1.0.0), downloading package again..."
                in caplog.text
            )

        with mock.patch.object(
            PackageManagerV1, "third_party_packages", new=packages
        ), mock.patch("shutil.rmtree"), caplog.at_level(logging.INFO):
            result = self.run_cli_command(
                "packages",
                "sync",
                "--update-hashes",
            )
            assert result.exit_code == 0
            assert (
                "Updating (protocol, open_aea/signing:1.0.0:bafybeiambqptflge33eemdhis2whik67hjplfnqwieoa6wblzlaf7vuo41)"
                in caplog.text
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
This module contain the test for the HTTP connection and skill documentation page.

The testing strategy is to just check that the Python code reported in
the documentation is contained in the http_echo skill.

Moreover, test that the code is compilable/executable.
"""

from pathlib import Path
from unittest import mock

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import BaseTestMarkdownDocs


class TestHttpConnectionAndSkill(BaseTestMarkdownDocs):
    """Test the skill testing code snippets."""

    DOC_PATH = Path(ROOT_DIR, "docs", "http-connection-and-skill.md")
    HTTP_ECHO_PATH = Path(ROOT_DIR, "packages", "fetchai", "skills", "http_echo")

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        super().setup_class()

        assert (
            len(cls.python_blocks) == 3
        ), "this test expected exactly 3 Python blocks."

        # read the content of 'handlers.py' and 'dialogues.py' of skill
        cls.http_echo_handlers_module = (cls.HTTP_ECHO_PATH / "handlers.py").read_text()
        cls.http_echo_dialogues = (cls.HTTP_ECHO_PATH / "dialogues.py").read_text()

    def test_handlers_code_snippet(self):
        """Test the 'handlers.py' code snippet."""
        handlers_code_snippet = self.python_blocks[0]["text"]

        # the handlers code snippet contains the YOUR_USERNAME placeholder
        # to be replaced by the user. We need to replace it with 'fetchai'
        # in order to compare it with the skill 'fetchai/http_echo'.
        replacement = "from packages.YOUR_USERNAME.skills.http_echo.dialogues import"
        to_be_replaced = "from packages.fetchai.skills.http_echo.dialogues import"
        expected_handlers_code = self.http_echo_handlers_module.replace(
            to_be_replaced, replacement
        )

        assert handlers_code_snippet in expected_handlers_code

    def test_dialogues_code_snippet(self):
        """Test the 'dialogues.py' code snippet."""
        dialogues_code_snippet = self.python_blocks[1]["text"]

        expected_dialogues_code = self.http_echo_dialogues
        assert dialogues_code_snippet in expected_dialogues_code

    @mock.patch("requests.get")
    @mock.patch("requests.post")
    def test_requests_code_snippet(self, *_mocks):
        """Test the 'requests' code snippet."""
        _globals, locals_ = {}, {}
        requests_code_snippet = self.python_blocks[2]["text"]
        exec(requests_code_snippet, _globals, locals_)  # nosec

        assert "requests" in locals_
        assert "response" in locals_



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains resilience and fault tolerance tests for P2PLibp2p connection."""

# pylint: skip-file

import pytest

from aea.multiplexer import Empty, Multiplexer
from aea.test_tools.utils import wait_for_condition

from packages.valory.connections.p2p_libp2p.check_dependencies import build_node
from packages.valory.connections.p2p_libp2p.tests.base import (
    SKIP_WINDOWS,
    libp2p_log_on_failure_all,
)
from packages.valory.connections.test_libp2p.tests.base import BaseP2PLibp2pTest

TIMEOUT = 10


@SKIP_WINDOWS
class BaseTestLibp2pRelay(BaseP2PLibp2pTest):
    """Base test class for libp2p connection relay."""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()
        build_node(cls.tmp)

    def change_state_and_wait(
        self,
        multiplexer: Multiplexer,
        expected_is_connected: bool = False,
        timeout: int = TIMEOUT,
    ) -> None:
        """Change state of a multiplexer (either connect or disconnect) and wait."""
        wait_for_condition(
            lambda: multiplexer.is_connected == expected_is_connected, timeout=timeout
        )


@libp2p_log_on_failure_all
class TestLibp2pConnectionRelayNodeRestart(BaseTestLibp2pRelay):
    """Test that connection will reliably forward envelopes after its relay node restarted"""

    def setup_method(self) -> None:
        """Set up the individual test method"""

        self.genesis = self.make_connection()
        genesis_peer = self.genesis.node.multiaddrs[0]
        self.relay = self.make_connection(entry_peers=[genesis_peer])
        relay_peer = self.relay.node.multiaddrs[0]
        self.connection1 = self.make_connection(relay=False, entry_peers=[relay_peer])
        self.connection2 = self.make_connection(relay=False, entry_peers=[relay_peer])

        # create references for disconnecting and reconnecting
        self.multiplexer_genesis = self.multiplexers[0]
        self.multiplexer_relay = self.multiplexers[1]
        self.multiplexer1 = self.multiplexers[2]
        self.multiplexer2 = self.multiplexers[3]

    def teardown_method(self) -> None:
        """Teardown"""
        self._disconnect()
        self.multiplexers.clear()

    def test_connection_is_established(self) -> None:
        """Test connection established."""
        assert self.all_connected

    def test_envelope_routed_from_peer_after_relay_restart(self) -> None:
        """Test envelope routed from third peer after relay restart."""

        to = self.connection1.address
        sender = self.genesis.address

        def attempt_sending() -> None:
            envelope = self.enveloped_default_message(to=to, sender=sender)
            self.multiplexer_genesis.put(envelope)
            delivered_envelope = self.multiplexer1.get(block=True, timeout=TIMEOUT)
            assert delivered_envelope is not None
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

        assert self.multiplexer_relay.is_connected
        attempt_sending()

        self.multiplexer_relay.disconnect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=False)
        assert not self.multiplexer_relay.is_connected

        with pytest.raises(Empty):
            attempt_sending()

        self.multiplexer_relay.connect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=True)
        assert self.multiplexer_relay.is_connected
        attempt_sending()

    def test_envelope_routed_from_client_after_relay_restart(self) -> None:
        """Test envelope routed from third relay client after relay restart."""

        to = self.connection1.address
        sender = self.connection2.address

        def attempt_sending() -> None:
            envelope = self.enveloped_default_message(to=to, sender=sender)
            self.multiplexer2.put(envelope)
            delivered_envelope = self.multiplexer1.get(block=True, timeout=TIMEOUT)
            assert delivered_envelope is not None
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

        assert self.multiplexer_relay.is_connected
        attempt_sending()

        self.multiplexer_relay.disconnect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=False)
        assert not self.multiplexer_relay.is_connected

        with pytest.raises(Empty):
            attempt_sending()

        self.multiplexer_relay.connect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=True)
        assert self.multiplexer_relay.is_connected
        attempt_sending()

    def test_envelope_routed_after_relay_restart(self) -> None:
        """Test envelope routed after relay restart."""

        sender = self.connection1.address
        to = self.genesis.address

        def attempt_sending() -> None:
            envelope = self.enveloped_default_message(to=to, sender=sender)
            self.multiplexer1.put(envelope)
            delivered_envelope = self.multiplexer_genesis.get(
                block=True, timeout=TIMEOUT
            )
            assert delivered_envelope is not None
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

        assert self.multiplexer_relay.is_connected
        attempt_sending()

        self.multiplexer_relay.disconnect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=False)
        assert not self.multiplexer_relay.is_connected

        with pytest.raises(Empty):
            attempt_sending()

        self.multiplexer_relay.connect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=True)
        assert self.multiplexer_relay.is_connected
        attempt_sending()


@libp2p_log_on_failure_all
class TestLibp2pConnectionAgentMobility(BaseTestLibp2pRelay):
    """Test that connection will correctly route envelope to destination that changed its peer"""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()

        cls.genesis = cls.make_connection()
        genesis_peer = cls.genesis.node.multiaddrs[0]
        cls.connection1 = cls.make_connection(entry_peers=[genesis_peer])
        cls.connection2 = cls.make_connection(
            entry_peers=[genesis_peer],
            agent_key=cls.default_crypto,
        )

        cls.multiplexer_genesis = cls.multiplexers[0]
        cls.multiplexer1 = cls.multiplexers[1]
        cls.multiplexer2 = cls.multiplexers[2]

    def test_envelope_routed_after_peer_changed(self) -> None:
        """Test envelope routed after peer changed."""

        sender = self.connection1.address
        to = self.connection2.address

        def attempt_sending() -> None:
            envelope = self.enveloped_default_message(to=to, sender=sender)
            self.multiplexer1.put(envelope)
            delivered_envelope = self.multiplexer2.get(block=True, timeout=TIMEOUT)
            assert delivered_envelope is not None
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

        assert self.multiplexer2.is_connected
        attempt_sending()

        self.multiplexer2.disconnect()
        self.change_state_and_wait(self.multiplexer2, expected_is_connected=False)
        assert not self.multiplexer2.is_connected

        with pytest.raises(Empty):
            attempt_sending()

        self.multiplexer2.connect()
        self.change_state_and_wait(self.multiplexer2, expected_is_connected=True)

        self.multiplexer2.connect()
        self.change_state_and_wait(self.multiplexer2, expected_is_connected=True)
        assert self.multiplexer2.is_connected
        attempt_sending()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This package contains the behaviour for the fipa dummy buyer skill."""

from typing import Any, cast

from aea.helpers.search.models import Constraint, ConstraintType, Query
from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.skills.fipa_dummy_buyer.dialogues import FipaDialogues


class FIPAInitializerBehaviour(TickerBehaviour):
    """Fipa buyer cfp initializer."""

    def __init__(self, **kwargs: Any) -> None:
        """Init fipa behaviour."""
        if "opponent_address" not in kwargs:
            raise ValueError("Opponent address has to be specified for behaviour!")
        self.opponent_address: str = cast(str, kwargs.pop("opponent_address", None))
        self.is_enabled: bool = True
        super().__init__(**kwargs)

    def setup(self) -> None:
        """Implement the setup for the behaviour."""

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """
        if not self.is_enabled:
            return
        dialogues = cast(FipaDialogues, self.context.fipa_dialogues)
        cfp_msg, _ = dialogues.create(
            counterparty=self.opponent_address,
            performative=FipaMessage.Performative.CFP,
            query=Query([Constraint("something", ConstraintType(">", 1))]),
        )
        self.context.outbox.put_message(cfp_msg)
        self.context.logger.info("CFP message sent.")

    def teardown(self) -> None:
        """Implement the task teardown."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test account module."""

from unittest import mock

import pytest
from aea_ledger_ethereum_hwi.account import HWIAccount, HWIErrorCodes
from aea_ledger_ethereum_hwi.exceptions import HWIError
from hexbytes import HexBytes
from ledgerwallet.client import CommException, LedgerClient

DEFAULT_ACCOUNT_RESPONSE = b"A\x01\xfb\x02\xc3\x7f\xdc7B\xd8\x975b[*0\xc4\xb0c-\x1f:P|\xc2}+0\x08\x94\x9f\xf9\xc3?\x9bihO\x03~c\xaekn\xda\xdd\xfaQMo\xcd\xa8\xa0\xb16T\x9c*\x8f\x8e\xc9d\x90F\xf0\xcc(2a5f50C28F04A04B4b9671c02B2fBd757cb51d23\x97\x01nA<e\x08\xd9\x84\x1b\x9a\xbf<My\\E\xb2h\x93\x82$\xf4t\x83\xe2'\xa1a|R\x90"
DEFAULT_TX_RESPONSE = b"\x00\x0e\x14a\n\xf5\x13\xdf\x91\xf0\xbaE\x80\xbb9\x87k:\xd2\x01\xbc\x87\x8a-4\xf9eDT\"$\x01\xda\x0b\xf6\x94\xeb\x1a->Z\xca\xa0\\e\x9e\xe9\xef\x82\xb9\xf8+\xac\xf5\xc6\x0b\x06P\xb2\xd9\xf03\xe7'\x13"
CHAIN_ID = 0
NONCE = 1
ETH_ACCOUNT = "0xe81DE7001292e482D4d1851FF7eD50c56093F8bb"
ETH_PUBLIC_KEY = "0xfb02c37fdc3742d89735625b2a30c4b0632d1f3a507cc27d2b3008949ff9c33f9b69684f037e63ae6b6edaddfa514d6fcda8a0b136549c2a8f8ec9649046f0cc"

enumerate_devices_patch = mock.patch(
    "aea_ledger_ethereum_hwi.account.enumerate_devices",
    return_value=[
        mock.MagicMock(),
    ],
)


def test_get_account() -> None:
    """Test Account.get_account"""

    with mock.patch.object(
        LedgerClient,
        "apdu_exchange",
        return_value=DEFAULT_ACCOUNT_RESPONSE,
    ), enumerate_devices_patch:
        account = HWIAccount().get_account()

    assert account.address == ETH_ACCOUNT
    assert account.public_key == ETH_PUBLIC_KEY


def test_sign_transaction_legacy() -> None:
    """Test sign legacy transaction"""

    account = HWIAccount()
    tx = {
        "chainId": CHAIN_ID,
        "nonce": NONCE,
        "to": ETH_ACCOUNT,
        "value": 1,
        "data": b"",
        "gas": 21000,
        "gasPrice": 1000000000,
    }

    with mock.patch.object(
        LedgerClient,
        "apdu_exchange",
        side_effect=[DEFAULT_ACCOUNT_RESPONSE, DEFAULT_TX_RESPONSE],
    ), enumerate_devices_patch:
        signed_tx = account.sign_transaction(transaction_dict=tx)

    assert signed_tx.raw_transaction == HexBytes(
        "0x01f8658001843b9aca0082520894e81de7001292e482d4d1851ff7ed50c56093f8bb0180c080a00e14610af513df91f0ba4580bb39876b3ad201bc878a2d34f9654454222401daa00bf694eb1a2d3e5acaa05c659ee9ef82b9f82bacf5c60b0650b2d9f033e72713"
    )
    assert signed_tx.hash == HexBytes(
        "0x0c6d3de9e1546f28d17a643c5dab1beb0354d33c3a91f046d30eb805a926009b"
    )
    assert (
        signed_tx.r
        == 6368386586266089348722015307950154351326471210129235368274039816995756835290
    )
    assert (
        signed_tx.s
        == 5411113509154936310743584815502120361180052194242088038643232853217736337171
    )
    assert signed_tx.v == 0


def test_sign_transaction_eip1559() -> None:
    """Test sign eip1559 transaction"""

    account = HWIAccount()
    tx = {
        "chainId": 5,
        "nonce": 7,
        "to": "0xf33BB476529e93Bb862262f6C6D8Cc324741d7aA",
        "value": 1,
        "data": b"",
        "gas": 21000,
        "maxFeePerGas": 170000000000,
        "maxPriorityFeePerGas": 3000000000,
    }

    with mock.patch.object(
        LedgerClient,
        "apdu_exchange",
        side_effect=[DEFAULT_ACCOUNT_RESPONSE, DEFAULT_TX_RESPONSE],
    ), enumerate_devices_patch:
        signed_tx = account.sign_transaction(transaction_dict=tx)

    assert signed_tx.raw_transaction == HexBytes(
        "0x02f86b050784b2d05e00852794ca240082520894f33bb476529e93bb862262f6c6d8cc324741d7aa0180c080a00e14610af513df91f0ba4580bb39876b3ad201bc878a2d34f9654454222401daa00bf694eb1a2d3e5acaa05c659ee9ef82b9f82bacf5c60b0650b2d9f033e72713"
    )
    assert signed_tx.hash == HexBytes(
        "0x5a65b647eb97bd910e8c6ba573503731adaa47213eb02c0c63316e727a46e64e"
    )
    assert (
        signed_tx.r
        == 6368386586266089348722015307950154351326471210129235368274039816995756835290
    )
    assert (
        signed_tx.s
        == 5411113509154936310743584815502120361180052194242088038643232853217736337171
    )
    assert signed_tx.v == 0


def test_exceptions() -> None:
    """Test exceptions."""

    with mock.patch.object(
        LedgerClient,
        "apdu_exchange",
        side_effect=CommException(message="0", sw=HWIErrorCodes.DEVICE_LOCKED),
    ), pytest.raises(HWIError, match="Cannot find any ledger device"):
        HWIAccount().get_account()

    with mock.patch.object(
        LedgerClient,
        "apdu_exchange",
        side_effect=CommException(message="0", sw=HWIErrorCodes.DEVICE_LOCKED),
    ), pytest.raises(HWIError, match="Device is locked"), mock.patch(
        "aea_ledger_ethereum_hwi.account.enumerate_devices",
        return_value=[
            mock.MagicMock(),
        ],
    ):
        HWIAccount().get_account()

    with mock.patch.object(
        LedgerClient,
        "apdu_exchange",
        side_effect=CommException(message="0", sw=HWIErrorCodes.EHTEREUM_APP_NOT_OPEN),
    ), pytest.raises(
        HWIError, match="Please open ethereum app in your device"
    ), mock.patch(
        "aea_ledger_ethereum_hwi.account.enumerate_devices",
        return_value=[
            mock.MagicMock(),
        ],
    ):
        HWIAccount().get_account()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the generic buyer skill."""

# pylint: skip-file

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

import pytest

from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.generic_buyer.behaviours import (
    GenericSearchBehaviour,
    GenericTransactionBehaviour,
    LEDGER_API_ADDRESS,
)
from packages.fetchai.skills.generic_buyer.dialogues import (
    FipaDialogue,
    FipaDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
)
from packages.fetchai.skills.generic_buyer.strategy import GenericStrategy
from packages.valory.connections.ledger.connection import PUBLIC_ID as LEDGER_PUBLIC_ID
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

ETHEREUM = "ethereum"
PACKAGE_ROOT = Path(__file__).parent.parent


class TestSearchBehaviour(BaseSkillTestCase):
    """Test Search behaviour of generic buyer."""

    path_to_skill = PACKAGE_ROOT

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.search_behaviour = cast(
            GenericSearchBehaviour, cls._skill.skill_context.behaviours.search
        )
        cls.tx_behaviour = cast(
            GenericTransactionBehaviour, cls._skill.skill_context.behaviours.transaction
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)

        cls.logger = cls._skill.skill_context.logger

    def setup_method(self):
        """Setup."""
        super().setup_method()
        self._init_strategy_kwargs = self.strategy.__dict__.copy()

    def teardown_method(self):
        """Teardown"""
        super().teardown_method()
        self.strategy.__dict__.update(self._init_strategy_kwargs)
        self.tx_behaviour.waiting.clear()

    def test_setup_is_ledger_tx(self):
        """Test the setup method of the search behaviour where is_ledger_tx is True."""
        # operation
        self.search_behaviour.setup()

        # after
        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            to=str(LEDGER_PUBLIC_ID),
            sender=str(self.skill.public_id),
            ledger_id=ETHEREUM,
            address=self.skill.skill_context.agent_address,
        )
        assert has_attributes, error_str

    def test_setup_not_is_ledger_tx(self):
        """Test the setup method of the search behaviour where is_ledger_tx is False."""
        # setup
        self.strategy._is_ledger_tx = False

        # before
        assert not self.strategy.is_searching

        # operation
        self.search_behaviour.setup()

        # after
        assert self.strategy.is_searching

    def test_act_is_searching(self):
        """Test the act method of the search behaviour where is_searching is True."""
        # setup
        self.strategy._is_searching = True
        assert self.strategy.is_searching

        # operation
        self.search_behaviour.act()

        # after
        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.public_id),
            query=self.skill.skill_context.strategy.get_location_and_service_query(),
        )
        assert has_attributes, error_str

    def test_act_not_is_searching(self):
        """Test the act method of the search behaviour where is_searching is False."""
        # setup
        self.strategy._is_searching = False

        # operation
        self.search_behaviour.act()

        # after
        self.assert_quantity_in_outbox(0)

    def test_act_remaining_transactions(self):
        """Test the act method of the search behaviour where remaining_transactions_count > 0."""
        # setup
        self.strategy._is_searching = True
        self.tx_behaviour.waiting = ["some_dialogue"]

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.search_behaviour.act()

        # after
        self.assert_quantity_in_outbox(0)
        mock_logger.assert_any_call(
            logging.INFO,
            f"Transaction behaviour has {len(self.tx_behaviour.waiting)} transactions remaining. Skipping search!",
        )

    def test_teardown(self):
        """Test the teardown method of the search behaviour."""
        assert self.search_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestTransactionBehaviour(BaseSkillTestCase):
    """Test transaction behaviour of generic buyer."""

    path_to_skill = PACKAGE_ROOT

    @classmethod
    def setup_class(cls):
        """Setup the test class."""

        super().setup_class()
        cls.transaction_behaviour = cast(
            GenericTransactionBehaviour, cls._skill.skill_context.behaviours.transaction
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)
        cls.logger = cls._skill.skill_context.logger

        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )

        cls.list_of_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
        )

    def setup_method(self):
        """Setup"""
        super().setup_method()
        self._init_strategy_kwargs = self.strategy.__dict__.copy()
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_messages,
            ),
        )
        fipa_dialogue.terms = "terms"  # type: ignore

        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=(
                    DialogueMessage(
                        LedgerApiMessage.Performative.GET_BALANCE,
                        {"ledger_id": "some_ledger_id", "address": "some_address"},
                    ),
                ),
            ),
        )
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue

        self.ledger_api_dialogue = ledger_api_dialogue
        self.fipa_dialogue = fipa_dialogue

    def teardown_method(self):
        """Teardown"""
        super().teardown_method()
        self.strategy.__dict__.update(self._init_strategy_kwargs)
        self.transaction_behaviour.teardown()

    def _check_start_processing_effects(self, fipa_dialogue, mock_logger) -> None:
        """Perform checks related to running _start_processing."""
        # _start_processing
        mock_logger.assert_any_call(
            logging.INFO,
            f"Processing transaction, {len(self.transaction_behaviour.waiting)} transactions remaining",
        )

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.public_id),
            terms=fipa_dialogue.terms,
        )
        assert has_attributes, error_str

        ledger_api_dialogue = cast(
            LedgerApiDialogue, self.ledger_api_dialogues.get_dialogue(message)
        )
        logging.error(">" * 100)
        logging.error(fipa_dialogue)
        logging.error(ledger_api_dialogue.associated_fipa_dialogue)
        assert ledger_api_dialogue.associated_fipa_dialogue == fipa_dialogue

        assert self.transaction_behaviour.processing_time == 0.0

        assert self.transaction_behaviour.processing == ledger_api_dialogue

        mock_logger.assert_any_call(
            logging.INFO,
            f"requesting transfer transaction from ledger api for message={message}...",
        )

    def test_setup(self):
        """Test the setup method of the transaction behaviour."""
        assert self.transaction_behaviour.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_act_i(self):
        """Test the act method of the transaction behaviour where processing IS None and len(self.waiting) is NOT 0."""
        # setup
        processing_time = 5.0
        max_processing = 120
        self.transaction_behaviour.max_processing = max_processing
        self.transaction_behaviour.processing_time = processing_time
        self.transaction_behaviour.waiting = [self.fipa_dialogue]

        # before
        assert self.transaction_behaviour.processing_time == processing_time
        assert self.transaction_behaviour.processing is None

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.transaction_behaviour.act()

        # after
        self.assert_quantity_in_outbox(1)

        # _start_processing
        self._check_start_processing_effects(self.fipa_dialogue, mock_logger)

    def test_act_ii(self):
        """Test the act method of the transaction behaviour where processing is NOT None and processing_time < max_processing."""
        # setup
        processing_time = 5.0
        self.transaction_behaviour.processing = "some_dialogue"  # TODO: type
        self.transaction_behaviour.max_processing = 120
        self.transaction_behaviour.processing_time = processing_time

        # operation
        self.transaction_behaviour.act()

        # after
        self.assert_quantity_in_outbox(0)
        assert (
            self.transaction_behaviour.processing_time
            == processing_time + self.transaction_behaviour.tick_interval
        )

    def test_act_iii(self):
        """Test the act method of the transaction behaviour where processing is NOT None and processing_time > max_processing."""
        # setup

        processing_time = 121.0
        self.transaction_behaviour.processing = self.ledger_api_dialogue
        self.transaction_behaviour.max_processing = 120
        self.transaction_behaviour.processing_time = processing_time

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.transaction_behaviour.act()

        # after
        self.assert_quantity_in_outbox(1)

        # _timeout_processing
        assert (
            self.ledger_api_dialogue.dialogue_label
            in self.transaction_behaviour.timedout
        )
        # below is overridden in _start_processing
        # assert self.fipa_dialogue in self.transaction_behaviour.waiting
        assert self.transaction_behaviour.processing_time == 0.0
        # below is overridden in _start_processing
        # assert self.transaction_behaviour.processing is None

        # _start_processing
        self._check_start_processing_effects(self.fipa_dialogue, mock_logger)

    def test_timeout_processing(self):
        """Test the _timeout_processing method of the transaction behaviour where self.processing IS None."""
        # setup
        self.transaction_behaviour.processing_time = None

        # operation
        self.transaction_behaviour._timeout_processing()

        # after
        self.assert_quantity_in_outbox(0)

    def test_act_iv(self):
        """Test the act method of the transaction behaviour where len(waiting) == 0."""
        # setup
        self.transaction_behaviour.waiting = []

        # operation
        self.transaction_behaviour.act()

        # after
        self.assert_quantity_in_outbox(0)

    def test_failed_processing(self):
        """Test the failed_processing method of the transaction behaviour."""
        # setup

        self.transaction_behaviour.timedout.add(self.ledger_api_dialogue.dialogue_label)

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.transaction_behaviour.failed_processing(self.ledger_api_dialogue)

        # after
        self.assert_quantity_in_outbox(0)

        # finish_processing
        assert self.transaction_behaviour.timedout == set()

        mock_logger.assert_any_call(
            logging.DEBUG,
            f"Timeout dialogue in transaction processing: {self.ledger_api_dialogue}",
        )

        # failed_processing
        assert self.fipa_dialogue in self.transaction_behaviour.waiting

    def test_finish_processing_i(self):
        """Test the finish_processing method of the transaction behaviour where self.processing == ledger_api_dialogue."""
        # setup

        self.transaction_behaviour.processing = self.ledger_api_dialogue

        # operation
        self.transaction_behaviour.failed_processing(self.ledger_api_dialogue)

        # after
        assert self.transaction_behaviour.processing_time == 0.0
        assert self.transaction_behaviour.processing is None

    def test_finish_processing_ii(self):
        """Test the finish_processing method of the transaction behaviour where ledger_api_dialogue's dialogue_label is NOT in self.timedout."""
        # setup

        # operation
        with pytest.raises(ValueError) as err:
            self.transaction_behaviour.finish_processing(self.ledger_api_dialogue)

        # after
        assert (
            err.value.args[0]
            == f"Non-matching dialogues in transaction behaviour: {self.transaction_behaviour.processing} and {self.ledger_api_dialogue}"
        )

    def test_teardown(self):
        """Test the teardown method of the transaction behaviour."""
        assert self.transaction_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)
